# encoding: utf-8

"""The matrix objects used by frozen-slice.

A matrix object has rows and (usually) columns.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from collections import namedtuple
import numpy as np
from scipy.stats.contingency import expected_freq

from cr.cube.enum import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty
from cr.cube.vector import (
    AssembledInsertionVector,
    AssembledVector,
    BaseVector,
    CategoricalVector,
    CatXMrVector,
    HiddenVector,
    MeansVector,
    MeansWithMrVector,
    MultipleResponseVector,
    OrderedVector,
    PrunedVector,
)


class _BaseTransformedMatrix(object):
    """Base class for late-stage matrices that transform a base matrix."""

    def __init__(self, base_slice, transforms):
        self._base_slice = base_slice
        self._transforms = transforms

    @lazyproperty
    def table_base(self):
        return self._base_slice.table_base

    @lazyproperty
    def table_margin(self):
        return self._base_slice.table_margin


class OrderedMatrix(_BaseTransformedMatrix):
    """Result of the ordering transform.

    In charge of indexing rows and columns properly.
    """

    @lazyproperty
    def columns(self):
        return tuple(
            OrderedVector(column, self._ordering.row_order)
            for column in tuple(
                np.array(self._base_slice.columns)[self._ordering.column_order]
            )
        )

    @lazyproperty
    def rows(self):
        return tuple(
            OrderedVector(row, self._ordering.column_order)
            for row in tuple(np.array(self._base_slice.rows)[self._ordering.row_order])
        )

    @lazyproperty
    def _ordering(self):
        return self._transforms.ordering


class PrunedMatrix(_BaseTransformedMatrix):
    """Matrix with rows or columns pruned.

    While the rows and/or columns need to be pruned, each one of the remaining
    vectors also needs to be pruned based on the opposite dimension's base.
    """

    @lazyproperty
    def columns(self):
        if not self._applied:
            return self._base_slice.columns

        return tuple(
            PrunedVector(column, self._base_slice.rows)
            for column in self._base_slice.columns
            if not column.pruned
        )

    @lazyproperty
    def rows(self):
        if not self._applied:
            return self._base_slice.rows

        return tuple(
            PrunedVector(row, self._base_slice.columns)
            for row in self._base_slice.rows
            if not row.pruned
        )

    @lazyproperty
    def table_base(self):
        if not self._applied:
            return self._base_slice.table_base

        margin = self._base_slice.table_base
        index = margin != 0
        if margin.ndim < 2:
            return margin[index]
        row_ind = np.any(index, axis=1)
        col_ind = np.any(index, axis=0)
        return margin[np.ix_(row_ind, col_ind)]

    @lazyproperty
    def table_base_unpruned(self):
        return self._base_slice.table_base

    @lazyproperty
    def table_margin(self):
        if not self._applied:
            return self._base_slice.table_margin

        margin = self._base_slice.table_margin
        index = margin != 0
        if margin.ndim < 2:
            return margin[index]
        row_ind = np.any(index, axis=1)
        col_ind = np.any(index, axis=0)
        return margin[np.ix_(row_ind, col_ind)]

    @lazyproperty
    def table_margin_unpruned(self):
        return self._base_slice.table_margin

    @lazyproperty
    def _applied(self):
        return self._transforms._pruning


class MatrixWithHidden(_BaseTransformedMatrix):
    @lazyproperty
    def columns(self):
        return tuple(
            HiddenVector(column, self._base_slice.rows)
            for column in self._base_slice.columns
            if not column.hidden
        )

    @lazyproperty
    def rows(self):
        return tuple(
            HiddenVector(row, self._base_slice.columns)
            for row in self._base_slice.rows
            if not row.hidden
        )


class MatrixWithInsertions(_BaseTransformedMatrix):
    """Represents slice with both normal and inserted bits."""

    @lazyproperty
    def columns(self):
        return tuple(
            self._top_columns + self._interleaved_columns + self._bottom_columns
        )

    @lazyproperty
    def rows(self):
        return tuple(self._top_rows + self._interleaved_rows + self._bottom_rows)

    @lazyproperty
    def _assembled_columns(self):
        return tuple(
            AssembledVector(column, self._insertion_rows)
            for column in self._base_slice.columns
        )

    @lazyproperty
    def _assembled_insertion_columns(self):
        return tuple(
            AssembledInsertionVector(column, self._insertion_rows)
            for column in self._insertions.columns
        )

    @lazyproperty
    def _assembled_insertion_rows(self):
        return tuple(
            AssembledInsertionVector(row, self._insertion_columns)
            for row in self._insertions.rows
        )

    @lazyproperty
    def _assembled_rows(self):
        return tuple(
            AssembledVector(row, self._insertion_columns)
            for row in self._base_slice.rows
        )

    @lazyproperty
    def _bottom_columns(self):
        return tuple(
            AssembledInsertionVector(column, self._insertion_rows)
            for column in self._insertions.bottom_columns
        )

    @lazyproperty
    def _bottom_rows(self):
        return tuple(
            AssembledInsertionVector(row, self._insertion_columns)
            for row in self._insertions.bottom_rows
        )

    @lazyproperty
    def _insertion_columns(self):
        return self._insertions._inserted_columns

    @lazyproperty
    def _insertion_rows(self):
        return self._insertions._rows

    @lazyproperty
    def _insertions(self):
        return self._transforms.insertions

    @lazyproperty
    def _interleaved_columns(self):
        columns = []
        for i in range(len(self._base_slice.columns)):
            columns.append(self._assembled_columns[i])
            for insertion_column in self._assembled_insertion_columns:
                if i == insertion_column.anchor:
                    columns.append(insertion_column)
        return tuple(columns)

    @lazyproperty
    def _interleaved_rows(self):
        rows = []
        for i in range(len(self._base_slice.rows)):
            rows.append(self._assembled_rows[i])
            for insertion_row in self._assembled_insertion_rows:
                if i == insertion_row.anchor:
                    rows.append(insertion_row)
        return tuple(rows)

    @lazyproperty
    def _top_rows(self):
        return tuple(
            AssembledVector(row, self._insertion_columns)
            for row in self._insertions.top_rows
        )

    @lazyproperty
    def _top_columns(self):
        return tuple(
            AssembledInsertionVector(column, self._insertion_rows)
            for column in self._insertions.top_columns
        )


# === pre-transform Matrix objects ===


# ---Used to represent the non-existent dimension in case of 1D vectors (that need to be
# ---accessed as slices, to support cr.exporter).
_PlaceholderElement = namedtuple("_PlaceholderElement", "label, is_hidden")


class MatrixFactory(object):
    """Encapsulates creation of the right raw (pre-transforms) matrix object."""

    @classmethod
    def matrix(
        cls,
        dimensions,
        counts,
        base_counts,
        counts_with_missings,
        cube,
        slice_idx,
        ca_as_0th,
    ):
        """Return a matrix object of appropriate type based on parameters."""

        # For cubes with means, cerate one of means-matrix types
        if cube.has_means:
            return cls._create_means_matrix(
                counts, base_counts, cube, dimensions, slice_idx
            )

        dimension_types = cube.dimension_types[-2:]
        if cube.ndim > 2 or ca_as_0th:
            base_counts = base_counts[slice_idx]
            counts = counts[slice_idx]
            counts_with_missings = counts_with_missings[slice_idx]
            if cube.dimension_types[0] == DT.MR:
                base_counts = base_counts[0]
                counts = counts[0]
                counts_with_missings = counts_with_missings[0]
            elif ca_as_0th:
                table_name = "%s: %s" % (
                    cube.dimensions[-2:][0].name,
                    cube.dimensions[-2:][0].valid_elements[slice_idx].label,
                )
                return _CaCatMatrix1D(dimensions, counts, base_counts, table_name)
        elif cube.ndim < 2:
            if dimension_types[0] == DT.MR:
                return _MrMatrix1D(dimensions, counts, base_counts)
            return _CatMatrix1D(dimensions, counts, base_counts)
        if dimension_types == (DT.MR, DT.MR):
            return _MrXMrMatrix(dimensions, counts, base_counts, counts_with_missings)
        elif dimension_types[0] == DT.MR:
            return _MrXCatMatrix(dimensions, counts, base_counts, counts_with_missings)
        elif dimension_types[1] == DT.MR:
            return _CatXMrMatrix(dimensions, counts, base_counts, counts_with_missings)
        return _CatXCatMatrix(dimensions, counts, base_counts, counts_with_missings)

    @classmethod
    def _create_means_matrix(cls, counts, base_counts, cube, dimensions, slice_idx):
        if cube.ndim == 0:
            return _MeansMatrix0D(counts, base_counts)
        elif cube.ndim == 1:
            if dimensions[0].dimension_type == DT.MR:
                return _MrWithMeansMatrix1D(dimensions[0], counts, base_counts)
            return _MeansMatrix1D(dimensions[0], counts, base_counts)
        elif cube.ndim >= 2:
            if cube.ndim == 3:
                base_counts = base_counts[slice_idx]
                counts = counts[slice_idx]
            if dimensions[0].dimension_type == DT.MR:
                return _MrXCatMeansMatrix(dimensions, counts, base_counts)
            return _CatXCatMeansMatrix(dimensions, counts, base_counts)


class _MeansMatrix0D(object):
    """Represents slices with means (and no counts)."""

    # TODO: We might need to have 2 of these, one for 0-D, and one for 1-D mean cubes
    def __init__(self, means, base_counts):
        self._means = means
        self._base_counts = base_counts

    @lazyproperty
    def means(self):
        return self._means

    @lazyproperty
    def table_base(self):
        # TODO: Check why we expect mean instead of the real base in this case.
        return self.means

    @lazyproperty
    def table_margin(self):
        return np.sum(self._base_counts)


class _MeansMatrix1D(_MeansMatrix0D):
    def __init__(self, dimension, means, base_counts):
        super(_MeansMatrix1D, self).__init__(means, base_counts)
        self._dimension = dimension

    @lazyproperty
    def columns(self):
        """A single vector that is used only for pruning Means slices."""
        return (
            BaseVector(_PlaceholderElement("Means Summary", False), self._base_counts),
        )

    @lazyproperty
    def rows(self):
        """Rows for Means slice, that enable iteration over labels.

        These vectors are not used for any computations. `means` is used for that,
        directly. However, for the wirng of the exporter, these mean slices need to
        support some additional API, such as labels. And for that, they need to
        support row iteration.
        """
        return tuple(
            MeansVector(element, base_counts, np.array([means]))
            for element, base_counts, means in zip(
                self._dimension.valid_elements, self._base_counts, self._means
            )
        )

    @lazyproperty
    def table_base(self):
        return np.sum(self._base_counts)


class _MrWithMeansMatrix1D(_MeansMatrix1D):
    @lazyproperty
    def rows(self):
        return tuple(
            MeansWithMrVector(element, base_counts, means)
            for element, base_counts, means in zip(
                self._dimension.valid_elements, self._base_counts, self._means
            )
        )


class _CatXCatMatrix(object):
    """Deals with CAT x CAT data.

    Delegates most functionality to vectors (rows or columns), but calculates some
    values by itself (like table_margin).

    This class (or its inheritants) must be instantiated as a starting point when
    dealing with slices. Other classes that represents various stages of
    transformations, need to repro a portion of this class' API (like iterating over
    rows or columns).
    """

    def __init__(self, dimensions, counts, base_counts, counts_with_missings=None):
        self._dimensions = dimensions
        self._counts = counts
        self._base_counts = base_counts
        self._all_counts = counts_with_missings

    @lazyproperty
    def columns(self):
        return tuple(
            CategoricalVector(counts, base_counts, element, self.table_margin, zscore)
            for counts, base_counts, element, zscore in self._column_generator
        )

    @lazyproperty
    def names(self):
        return tuple([dimension.name for dimension in self._dimensions])

    @lazyproperty
    def rows(self):
        return tuple(
            CategoricalVector(
                counts, base_counts, element, self.table_margin, zscore, column_index
            )
            for (
                counts,
                base_counts,
                element,
                zscore,
                column_index,
            ) in self._row_generator
        )

    @lazyproperty
    def table_base(self):
        return np.sum(self._base_counts)

    @lazyproperty
    def table_margin(self):
        return np.sum(self._counts)

    @lazyproperty
    def _baseline(self):
        """ndarray of baseline values for column index.

        Baseline is obtained by calculating the unconditional row margin (which needs
        to include missings from the column dimension) and dividint it by the total.
        Total also needs to include the missings from the column dimension.

        `dim_sum` is the unconditional row margin. For CAT x CAT slice it needs to
        sum across axis 1, because that's the columns CAT dimension, that gets
        collapsed. The total is calculated by totaling the unconditional row margin.
        Please note that the total _doesn't_ include missings for the row dimension.
        """
        dim_sum = np.sum(self._all_counts, axis=1)[self._valid_rows_idxs]
        return dim_sum[:, None] / np.sum(dim_sum)

    @lazyproperty
    def _column_dimension(self):
        return self._dimensions[1]

    @lazyproperty
    def _column_elements(self):
        return self._column_dimension.valid_elements

    @lazyproperty
    def _column_index(self):
        # TODO: This is a hack to make it work. It should be addressed properly with
        # passing `counts_with_missings` in all the right places in the factory.
        # Also - subclass for proper functionality in various MR cases.
        if self._all_counts is None:
            return self._column_proportions

        return self._column_proportions / self._baseline * 100

    @lazyproperty
    def _column_generator(self):
        return zip(
            self._counts.T, self._base_counts.T, self._column_elements, self._zscores.T
        )

    @lazyproperty
    def _column_proportions(self):
        return np.array([col.proportions for col in self.columns]).T

    @lazyproperty
    def _row_dimension(self):
        return self._dimensions[0]

    @lazyproperty
    def _row_elements(self):
        return self._row_dimension.valid_elements

    @lazyproperty
    def _row_generator(self):
        return zip(
            self._counts,
            self._base_counts,
            self._row_elements,
            self._zscores,
            self._column_index,
        )

    @staticmethod
    def _scalar_type_std_res(counts, total, colsum, rowsum):
        """Return ndarray containing standard residuals for category values.

        The shape of the return value is the same as that of *counts*.
        """
        expected_counts = expected_freq(counts)
        residuals = counts - expected_counts
        variance = (
            np.outer(rowsum, colsum)
            * np.outer(total - rowsum, total - colsum)
            / total ** 3
        )
        return residuals / np.sqrt(variance)

    @lazyproperty
    def _valid_rows_idxs(self):
        """ndarray-style index for only valid rows (out of missing and not-missing)."""
        return np.ix_(self._dimensions[-2].valid_elements.element_idxs)

    @lazyproperty
    def _zscores(self):
        return self._scalar_type_std_res(
            self._counts,
            self.table_margin,
            np.sum(self._counts, axis=0),
            np.sum(self._counts, axis=1),
        )


class _CatXCatMeansMatrix(_CatXCatMatrix):
    def __init__(self, dimensions, means, base_counts):
        super(_CatXCatMeansMatrix, self).__init__(dimensions, None, base_counts)
        self._means = means

    @lazyproperty
    def columns(self):
        return tuple(
            MeansVector(element, base_counts, means)
            for element, base_counts, means in zip(
                self._column_dimension.valid_elements,
                self._base_counts.T,
                self._means.T,
            )
        )

    @lazyproperty
    def rows(self):
        return tuple(
            MeansVector(element, base_counts, means)
            for element, base_counts, means in zip(
                self._row_dimension.valid_elements, self._base_counts, self._means
            )
        )


class _CatMatrix1D(_CatXCatMatrix):
    """Special case of CAT x CAT, where the 2nd CAT doesn't exist.

    Values are treated as rows, while there's only a single column (vector).
    """

    @lazyproperty
    def columns(self):
        return tuple(
            [
                CategoricalVector(
                    self._counts,
                    self._base_counts,
                    _PlaceholderElement("Summary", False),
                    self.table_margin,
                )
            ]
        )

    @lazyproperty
    def _zscores(self):
        # TODO: Fix with real zscores
        return tuple([np.nan for _ in self._counts])


class _CaCatMatrix1D(_CatMatrix1D):
    def __init__(self, dimensions, counts, base_counts, table_name):
        super(_CaCatMatrix1D, self).__init__(dimensions, counts, base_counts)
        self._table_name = table_name


class _MatrixWithMR(_CatXCatMatrix):
    @staticmethod
    def _array_type_std_res(counts, total, colsum, rowsum):
        expected_counts = rowsum * colsum / total
        variance = rowsum * colsum * (total - rowsum) * (total - colsum) / total ** 3
        return (counts - expected_counts) / np.sqrt(variance)


class _MrXCatMatrix(_MatrixWithMR):
    """Represents MR x CAT slices.

    It's similar to CAT x CAT, other than the way it handles columns. For
    columns - which correspond to the MR dimension - it needs to handle the indexing
    of selected/not-selected correctly.
    """

    @lazyproperty
    def columns(self):
        """Use bother selected and not-selected counts."""
        return tuple(
            MultipleResponseVector(
                counts, base_counts, element, self.table_margin, zscore
            )
            for counts, base_counts, element, zscore in self._column_generator
        )

    @lazyproperty
    def rows(self):
        """Use only selected counts."""
        return tuple(
            CatXMrVector(
                counts, base_counts, element, table_margin, zscore, column_index
            )
            for (
                counts,
                base_counts,
                element,
                table_margin,
                zscore,
                column_index,
            ) in self._row_generator
        )

    @lazyproperty
    def table_base(self):
        return np.sum(self._base_counts, axis=(1, 2))

    @lazyproperty
    def table_margin(self):
        return np.sum(self._counts, axis=(1, 2))

    @lazyproperty
    def _baseline(self):
        """ndarray of baseline values for column index.

        Baseline is obtained by calculating the unconditional row margin (which needs
        to include missings from the column dimension) and dividint it by the total.
        Total also needs to include the missings from the column dimension.

        `dim_sum` is the unconditional row margin. For MR x CAT slice it needs to
        sum across axis 2, because that's the CAT dimension, that gets collapsed.
        Then it needs to select only the selected counts of
        the MR (hence the `[:, 0]` index). The total needs include missings of the
        2nd dimension, but not of the first (hence the [:, 0:2] index, which only
        includes the selected and not-selected of the MR dimension).
        """
        dim_sum = np.sum(self._all_counts, axis=2)[:, 0][self._valid_rows_idxs]
        total = np.sum(self._all_counts[self._valid_rows_idxs][:, 0:2], axis=(1, 2))
        return (dim_sum / total)[:, None]

    @lazyproperty
    def _row_generator(self):
        return zip(
            self._counts,
            self._base_counts,
            self._row_elements,
            self.table_margin,
            self._zscores,
            self._column_index,
        )

    @lazyproperty
    def _zscores(self):
        return self._array_type_std_res(
            self._counts[:, 0, :],
            self.table_margin[:, None],
            np.sum(self._counts, axis=1),
            np.sum(self._counts[:, 0, :], axis=1)[:, None],
        )


class _MrXCatMeansMatrix(_MrXCatMatrix):
    def __init__(self, dimensions, means, base_counts):
        counts = np.empty(means.shape)
        super(_MrXCatMeansMatrix, self).__init__(dimensions, counts, base_counts)
        self._means = means

    @lazyproperty
    def rows(self):
        return tuple(
            MeansWithMrVector(element, base_counts, means[0])
            for element, base_counts, means in zip(
                self._row_dimension.valid_elements, self._base_counts, self._means
            )
        )


class _MrMatrix1D(_MrXCatMatrix):
    """Special case of 1-D MR slice (vector)."""

    @lazyproperty
    def columns(self):
        return tuple(
            [
                MultipleResponseVector(
                    self._counts.T,
                    self._base_counts.T,
                    _PlaceholderElement("Summary", False),
                    self.table_margin,
                )
            ]
        )

    @lazyproperty
    def table_base(self):
        return np.sum(self._base_counts, axis=1)

    @lazyproperty
    def table_margin(self):
        return np.sum(self._counts, axis=1)

    @lazyproperty
    def _zscores(self):
        return np.array([np.nan] * self._base_counts.shape[0])


class _CatXMrMatrix(_MatrixWithMR):
    """Handles CAT x MR slices.

    Needs to handle correctly the indexing for the selected/not-selected for rows
    (which correspond to the MR dimension).
    """

    @lazyproperty
    def columns(self):
        return tuple(
            CatXMrVector(counts.T, base_counts.T, element, table_margin)
            for counts, base_counts, element, table_margin in self._column_generator
        )

    @lazyproperty
    def rows(self):
        return tuple(
            MultipleResponseVector(
                counts.T,
                base_counts.T,
                element,
                self.table_margin,
                zscore,
                column_index,
            )
            for (
                counts,
                base_counts,
                element,
                zscore,
                column_index,
            ) in self._row_generator
        )

    @lazyproperty
    def table_base(self):
        return np.sum(self._base_counts, axis=(0, 2))

    @lazyproperty
    def table_margin(self):
        return np.sum(self._counts, axis=(0, 2))

    @lazyproperty
    def _baseline(self):
        """ndarray of baseline values for column index.

        Baseline is obtained by calculating the unconditional row margin (which needs
        to include missings from the column dimension) and dividint it by the total.
        Total also needs to include the missings from the column dimension.

        `dim_sum` is the unconditional row margin. For CAT x MR slice it needs to sum
        across axis 2, because that's the MR_CAT dimension, that gets collapsed
        (but the MR subvars don't get collapsed). Then it needs to calculate the total,
        which is easily obtained by summing across the CAT dimension (hence `axis=0`).
        """
        dim_sum = np.sum(self._all_counts, axis=2)[self._valid_rows_idxs]
        return dim_sum / np.sum(dim_sum, axis=0)

    @lazyproperty
    def _column_generator(self):
        return zip(
            # self._counts.T[0],
            np.array([self._counts.T[0].T, self._counts.T[1].T]).T,
            # self._base_counts.T[0],
            np.array([self._base_counts.T[0].T, self._base_counts.T[1].T]).T,
            self._column_elements,
            self.table_margin,
        )

    @lazyproperty
    def _zscores(self):
        return self._array_type_std_res(
            self._counts[:, :, 0],
            self.table_margin,
            np.sum(self._counts[:, :, 0], axis=0),
            np.sum(self._counts, axis=2),
        )


class _MrXMrMatrix(_MatrixWithMR):
    """Represents MR x MR slices.

    Needs to properly index both rows and columns (selected/not-selected.
    """

    @lazyproperty
    def columns(self):
        return tuple(
            MultipleResponseVector(counts, base_counts, element, table_margin)
            for counts, base_counts, element, table_margin in self._column_generator
        )

    @lazyproperty
    def rows(self):
        return tuple(
            MultipleResponseVector(
                counts[0].T,
                base_counts[0].T,
                element,
                table_margin,
                zscore,
                column_index,
            )
            for (
                counts,
                base_counts,
                element,
                table_margin,
                zscore,
                column_index,
            ) in self._row_generator
        )

    @lazyproperty
    def table_base(self):
        return np.sum(self._base_counts, axis=(1, 3))

    @lazyproperty
    def table_margin(self):
        return np.sum(self._counts, axis=(1, 3))

    @lazyproperty
    def _baseline(self):
        """ndarray of baseline values for column index.

        Baseline is obtained by calculating the unconditional row margin (which needs
        to include missings from the column dimension) and dividint it by the total.
        Total also needs to include the missings from the column dimension.

        `dim_sum` is the unconditional row margin. For MR x MR slice it needs to sum
        across axis 3, because that's the MR_CAT dimension, that gets collapsed
        (but the MR subvars don't get collapsed). Then it needs to calculate the total,
        which is obtained by summing across both MR_CAT dimensions. However, please
        note, that in calculating the unconditional total, missing elements need to be
        included for the column dimension, while they need to be _excluded_ for the row
        dimension. Hence the `[:, 0:2]` indexing for the first MR_CAT, but not the 2nd.
        """
        dim_sum = np.sum(self._all_counts[:, 0:2], axis=3)[self._valid_rows_idxs][:, 0]
        total = np.sum(self._all_counts[:, 0:2], axis=(1, 3))[self._valid_rows_idxs]
        return dim_sum / total

    @lazyproperty
    def _column_generator(self):
        return zip(
            self._counts.T[0],
            self._base_counts.T[0],
            self._column_elements,
            self.table_margin.T,
        )

    @lazyproperty
    def _row_generator(self):
        return zip(
            self._counts,
            self._base_counts,
            self._row_elements,
            self.table_margin,
            self._zscores,
            self._column_index,
        )

    @lazyproperty
    def _zscores(self):
        return self._array_type_std_res(
            self._counts[:, 0, :, 0],
            self.table_margin,
            np.sum(self._counts, axis=1)[:, :, 0],
            np.sum(self._counts, axis=3)[:, 0, :],
        )



# See http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages
try:  # pragma: no cover
    __import__("pkg_resources").declare_namespace(__name__)
except ImportError:  # pragma: no cover
    from pkgutil import extend_path

    __path__ = extend_path(__path__, __name__)



# encoding: utf-8

"""Unit test suite for the  cr.cube.measures.scale_means module."""

from __future__ import division

import numpy as np
import pytest

from cr.cube.measures.scale_means import ScaleMeans

from ..unitutil import Mock, property_mock

# pylint: disable=missing-docstring, redefined-outer-name, protected-access

COLS_DIM_VALUES = np.array([0, 1, 2])
ROWS_DIM_VALUES = np.array([3, 4, 5, 6])
ROW_MARGIN = np.array([0, 1, 2])
COL_MARGIN = np.array([3, 4, 5, 6])


def margin(axis):
    return ROW_MARGIN if axis else COL_MARGIN


def test_scale_means_marginal(scale_means_fixture):
    scale_means, axis, expected = scale_means_fixture
    assert scale_means.margin(axis) == expected


def test_inflate(inflate_fixture):
    dim_ind, scale_means, expected = inflate_fixture
    assert scale_means._inflate(dim_ind) == expected


@pytest.fixture(
    params=[
        (
            [None, ROWS_DIM_VALUES],
            margin,
            0,
            2,
            np.sum(COL_MARGIN * ROWS_DIM_VALUES) / np.sum(COL_MARGIN),
        ),
        (
            [COLS_DIM_VALUES, None],
            margin,
            1,
            2,
            np.sum(ROW_MARGIN * COLS_DIM_VALUES) / np.sum(ROW_MARGIN),
        ),
        (
            [None, ROWS_DIM_VALUES],
            margin,
            0,
            2,
            np.sum(COL_MARGIN * ROWS_DIM_VALUES) / np.sum(COL_MARGIN),
        ),
        ([None, None], margin, 0, 2, None),
    ]
)
def scale_means_fixture(request, values_prop_):
    values, margin, axis, ndim, expected = request.param
    values_prop_.return_value = values
    slice_ = Mock()
    slice_.ndim = ndim
    slice_.margin = margin
    scale_means = ScaleMeans(slice_)
    return scale_means, axis, expected


@pytest.fixture
def values_prop_(request):
    return property_mock(request, ScaleMeans, "values")


@pytest.fixture(
    params=[
        (0, 2, (Mock(), Mock()), True),
        (1, 2, (Mock(), Mock()), False),
        (0, 2, (Mock(),), False),
        (0, 1, (Mock(),), False),
    ]
)
def inflate_fixture(request):
    dim_ind, ndim, shape, expected = request.param
    slice_ = Mock()
    slice_.ndim = ndim
    slice_.get_shape.return_value = shape
    scale_means = ScaleMeans(slice_)
    return dim_ind, scale_means, expected


def test_valid_indices(valid_indices_fixture):
    scale_means, axis, expected = valid_indices_fixture
    actual = scale_means._valid_indices(axis)
    np.testing.assert_equal(actual, expected)


@pytest.fixture(
    params=[
        ([Mock(numeric_values=[])], 0, [slice(None)]),
        ([Mock(numeric_values=[1, 2, 3])], 0, [np.array([True, True, True])]),
        (
            [Mock(numeric_values=[1, 2, np.nan, 4])],
            0,
            [np.array([True, True, False, True])],
        ),
        ([Mock(numeric_values=[1])], 0, [np.array([True])]),
        (
            [Mock(numeric_values=[1, 2, 3]), Mock(numeric_values=[])],
            0,
            [np.array([True, True, True])],
        ),
        ([Mock(numeric_values=[1, 2, 3]), Mock(numeric_values=[])], 1, [slice(None)]),
    ]
)
def valid_indices_fixture(request):
    dimensions, axis, expected = request.param
    slice_ = Mock()
    slice_.dimensions = dimensions
    return ScaleMeans(slice_), axis, expected



# encoding: utf-8

"""Integration test suite for the cr.cube.dimension module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pytest

from unittest import TestCase

from cr.cube.crunch_cube import CrunchCube
from cr.cube.dimension import (
    AllDimensions,
    Dimension,
    _NewAllElements,
    NewDimension,
    _NewElement,
    _Subtotal,
)
from cr.cube.enum import DIMENSION_TYPE as DT

from ..fixtures import CR  # ---mnemonic: CR = 'cube-response'---


class DescribeIntegratedAllDimensions(object):
    """Integration-test suite for `cr.cube.dimension.AllDimensions` object."""

    def it_resolves_the_type_of_each_dimension(self, type_fixture):
        dimension_dicts, expected_types = type_fixture
        all_dimensions = AllDimensions(dimension_dicts)

        dimension_types = tuple(d.dimension_type for d in all_dimensions)

        assert dimension_types == expected_types

    def it_provides_access_to_the_apparent_dimensions(self):
        dimension_dicts = CR.CA_X_MR_WEIGHTED_HS["result"]["dimensions"]
        all_dimensions = AllDimensions(dimension_dicts)

        apparent_dimension_types = tuple(
            d.dimension_type for d in all_dimensions.apparent_dimensions
        )

        assert apparent_dimension_types == (DT.CA, DT.CA_CAT, DT.MR)

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            (CR.CAT_X_CAT, (DT.CAT, DT.CAT)),
            (CR.CA_X_MR_WEIGHTED_HS, (DT.CA, DT.CA_CAT, DT.MR, DT.MR_CAT)),
        ]
    )
    def type_fixture(self, request):
        cube_response, expected_types = request.param
        dimension_dicts = cube_response["result"]["dimensions"]
        return dimension_dicts, expected_types


class DescribeIntegratedNewDimension(object):
    """Integration-test suite for `cr.cube.dimension.NewDimension` object.

    This is a temporary object, roughly a shim, to provide the new Dimension interface
    to FrozenSlice. Once that work is completed, this will become DescribeDimension and
    the NewDimension object will be merged into the Dimension object.
    """

    def it_constructs_from_a_legacy_Dimension_object(self, legacy_dimension):
        dimension = NewDimension(legacy_dimension, None)
        assert isinstance(dimension, NewDimension)

    def it_provides_access_to_all_elements_in_its_collection(self, legacy_dimension):
        dimension_transforms_dict = {}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        elements = dimension.all_elements

        assert isinstance(elements, _NewAllElements)

    def it_knows_its_transformed_description(self, legacy_dimension):
        """Including that it resolves the dimension-rename transform cascade."""
        dimension_transforms_dict = {"description": "foobar"}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        description = dimension.description

        assert description == "foobar"

    def but_it_uses_element_description_when_not_transformed(self, legacy_dimension):
        """Including that it resolves the dimension-rename transform cascade."""
        dimension_transforms_dict = {}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        description = dimension.description

        assert description == (
            "If President Obama and the Republicans in Congress do not reach a budget"
            " agreement in time to avoid a shutdown of the federal government, who do"
            " you think will more to blame--President Obama or the Republican Congres"
            "s?"
        )

    def it_knows_the_display_order_of_its_elements(self, legacy_dimension):
        dimension_transforms_dict = {
            "order": {"type": "explicit", "element_ids": [3, 5, 42, 8, 2]}
        }
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        display_order = dimension.display_order

        assert display_order == (2, 4, 5, 1, 0, 3, 6, 7)

    def it_knows_its_transformed_name(self, legacy_dimension):
        dimension_transforms_dict = {"name": "barfoo"}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        name = dimension.name

        assert name == "barfoo"

    def but_it_uses_element_name_when_not_transformed(self, legacy_dimension):
        dimension_transforms_dict = {}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        name = dimension.name

        assert name == "ShutdownBlame"

    def it_knows_whether_it_should_be_pruned(self, legacy_dimension):
        dimension_transforms_dict = {"prune": True}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        prune = dimension.prune

        assert prune is True

    def it_provides_access_to_its_inserted_subtotal_specs(self, legacy_dimension):
        dimension_transforms_dict = {}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        subtotals = dimension.subtotals

        assert len(subtotals) == 1

    def but_it_uses_transforms_insertions_instead_when_present(self, legacy_dimension):
        dimension_transforms_dict = {"insertions": []}
        dimension = NewDimension(legacy_dimension, dimension_transforms_dict)

        subtotals = dimension.subtotals

        assert len(subtotals) == 0

    # fixture components ---------------------------------------------

    @pytest.fixture
    def legacy_dimension(self):
        return CrunchCube(CR.ECON_BLAME_WITH_HS).dimensions[0]


class DescribeIntegrated_NewAllElements(object):
    """Integration-test suite for `cr.cube.dimension._NewAllElements` object.

    _NewAllElements is a temporary object, roughly a shim, to provide the new
    _AllElements interface to FrozenSlice. Once that work is completed, this will become
    DescribeIntegrated_AllElements and the _NewAllElements object will be merged into
    the _AllElements object.
    """

    def it_knows_the_transformed_element_display_order(self, type_dict):
        dimension_transforms_dict = {
            "order": {"type": "explicit", "element_ids": [2, 1, 666, 4, 3, 4, 8]}
        }
        all_elements = _NewAllElements(type_dict, dimension_transforms_dict)

        display_order = all_elements.display_order

        assert display_order == (1, 0, 3, 2, 5, 4, 6, 7)
        assert len(display_order) == len(all_elements)

    def but_it_returns_the_default_display_order_when_not_transformed(self, type_dict):
        dimension_transforms_dict = {}
        all_elements = _NewAllElements(type_dict, dimension_transforms_dict)

        order = all_elements.display_order

        assert order == (0, 1, 2, 3, 4, 5, 6, 7)

    def it_constructs_its_element_objects_to_help(self, type_dict):
        dimension_transforms_dict = {}
        all_elements = _NewAllElements(type_dict, dimension_transforms_dict)

        elements = all_elements._elements

        assert all(isinstance(element, _NewElement) for element in elements)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def type_dict(self):
        return CrunchCube(CR.ECON_BLAME_WITH_HS).dimensions[0]._dimension_dict["type"]


class DescribeIntegrated_NewElement(object):
    """Integration-test suite for `cr.cube.dimension._NewElement` object.

    _NewElement is a temporary object, roughly a shim, to provide the new _Element
    interface to FrozenSlice. Once work on FrozenSlice is completed, this will become
    DescribeIntegrated_Element and the _NewElement object will be merged into the
    _BaseElement object and become _Element. Distinct _BaseElement, and _Category
    classes will go away; only _Element will remain.
    """

    def it_knows_its_transformed_label(self, element_dict):
        element_transforms_dict = {"name": "Xfinity Lounge"}
        element = _NewElement(element_dict, None, None, element_transforms_dict)

        label = element.label

        assert label == "Xfinity Lounge"

    def but_it_uses_its_base_name_if_no_transform_is_present(self, element_dict):
        element_transforms_dict = {}
        element = _NewElement(element_dict, None, None, element_transforms_dict)

        label = element.label

        assert label == "President Obama"

    def it_knows_when_it_is_explicitly_hidden(self, element_dict):
        element_transforms_dict = {"hide": True}
        element = _NewElement(element_dict, None, None, element_transforms_dict)

        is_hidden = element.is_hidden

        assert is_hidden is True

    def but_it_is_not_hidden_by_default(self):
        element_transforms_dict = {}
        element = _NewElement(None, None, None, element_transforms_dict)

        is_hidden = element.is_hidden

        assert is_hidden is False

    # fixture components ---------------------------------------------

    @pytest.fixture
    def element_dict(self):
        return (
            CrunchCube(CR.ECON_BLAME_WITH_HS)
            .dimensions[0]
            ._dimension_dict["type"]["categories"][0]
        )


class TestDimension(TestCase):
    """Legacy integration-test suite for Dimension object."""

    def test_subtotals_indices_single_subtotal(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS).dimensions[0]
        hs_indices = dimension.hs_indices
        self.assertEqual(hs_indices, ((1, (0, 1)),))

    def test_inserted_hs_indices_single_subtotal(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS).dimensions[0]
        # It can be verified what the inserted indices are, by comparing
        # labels with/without transforms.
        expected = [2]
        actual = dimension.inserted_hs_indices
        assert actual == expected

    def test_labels_for_categoricals(self):
        dimension_dict = {
            "type": {
                "class": "categorical",
                "categories": [
                    {"id": 1, "name": "Cat", "missing": False},
                    {"id": 2, "name": "Mouse", "missing": False},
                    {"id": -1, "name": "Iguana", "missing": True},
                ],
            }
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        # ---get only non-missing---
        labels = dimension.labels()
        assert labels == ["Cat", "Mouse"]

        # ---get all---
        labels = dimension.labels(include_missing=True)
        assert labels == ["Cat", "Mouse", "Iguana"]

    def test_labels_for_numericals(self):
        dimension_dict = {
            "type": {
                "class": "enum",
                "elements": [
                    {"id": 0, "value": "smallish", "missing": False},
                    {"id": 1, "value": "kinda big", "missing": False},
                    {"id": 2, "value": {}, "missing": True},
                ],
                "subtype": {"class": "numeric"},
            }
        }
        dimension = Dimension(dimension_dict, DT.BINNED_NUMERIC)

        # ---non-missing labels---
        labels = dimension.labels()
        assert labels == ["smallish", "kinda big"]

        # ---all labels, both valid and missing---
        labels = dimension.labels(include_missing=True)
        assert labels == ["smallish", "kinda big", ""]

        # ---all labels, both valid and missing---
        labels = dimension.labels(include_cat_ids=True)
        assert labels == [("smallish", 0), ("kinda big", 1)]

    def test_subtotals_indices_two_subtotals(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING).dimensions[0]
        hs_indices = dimension.hs_indices
        self.assertEqual(hs_indices, ((1, (0, 1)), ("bottom", (3, 4))))

    def test_inserted_hs_indices_two_subtotals(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING).dimensions[0]
        # It can be verified what the inserted indices are, by comparing
        # labels with/without transforms.
        expected = [2, 6]
        actual = dimension.inserted_hs_indices
        self.assertEqual(actual, expected)

    def test_inserted_hs_indices_order_and_labels(self):
        dimension_dict = {
            "references": {
                "view": {
                    "transform": {
                        "insertions": [
                            {
                                "anchor": "bottom",
                                "args": [111],
                                "function": "subtotal",
                                "name": "bottoms up one",
                            },
                            {
                                "anchor": "bottom",
                                "args": [222],
                                "function": "subtotal",
                                "name": "bottoms up two",
                            },
                            {
                                "anchor": "bottom",
                                "args": [333],
                                "function": "subtotal",
                                "name": "bottoms up three",
                            },
                            {
                                "anchor": "top",
                                "args": [444],
                                "function": "subtotal",
                                "name": "on top one",
                            },
                            {
                                "anchor": "top",
                                "args": [555],
                                "function": "subtotal",
                                "name": "on top two",
                            },
                            {
                                "anchor": 333,
                                "args": [555],
                                "function": "subtotal",
                                "name": "in the middle one",
                            },
                            {
                                "anchor": 333,
                                "args": [555],
                                "function": "subtotal",
                                "name": "in the middle two",
                            },
                        ]
                    }
                }
            },
            "type": {
                "categories": [
                    {"id": 111},
                    {"id": 222},
                    {"id": 333},
                    {"id": 444},
                    {"id": 555},
                ],
                "class": "categorical",
            },
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        assert dimension.inserted_hs_indices == [0, 1, 5, 6, 9, 10, 11]
        assert dimension.labels(include_transforms=True) == [
            "on top one",
            "on top two",
            "",
            "",
            "",
            "in the middle one",
            "in the middle two",
            "",
            "",
            "bottoms up one",
            "bottoms up two",
            "bottoms up three",
        ]

    def test_has_transforms_false(self):
        dimension = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).dimensions[1]
        expected = False
        actual = dimension.has_transforms
        self.assertEqual(actual, expected)

    def test_has_transforms_true(self):
        dimension = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).dimensions[0]
        expected = True
        actual = dimension.has_transforms
        self.assertEqual(actual, expected)

    def test_hs_indices_for_mr(self):
        dimension = CrunchCube(CR.MR_X_CAT_HS)._all_dimensions[1]
        hs_indices = dimension.hs_indices
        assert hs_indices == ()

    def test_hs_indices_with_bad_data(self):
        cube = CrunchCube(CR.CA_WITH_NETS)

        subvar_dim = cube.dimensions[0]
        anchor_idxs = [anchor_idx for anchor_idx, _ in subvar_dim.hs_indices]
        assert anchor_idxs == ["bottom", "bottom"]

        cat_dim = cube.dimensions[1]
        anchor_idxs = [anchor_idx for anchor_idx, _ in cat_dim.hs_indices]
        assert anchor_idxs == ["bottom", "bottom"]

    def test_skips_bad_data_for_hs_indices(self):
        """Test H&S indices with bad input data.

        This test ensures that H&S functionality doesn't break if it
        encounters bad transformations data, as is possible with some of the
        leftovers in the variables.
        """
        dimension_dict = {
            "references": {
                "view": {
                    "transform": {
                        "insertions": [
                            {"anchor": 101, "name": "This is respondent ideology"},
                            {
                                "anchor": 2,
                                "args": [1, 2],
                                "function": "subtotal",
                                "name": "Liberal net",
                            },
                            {
                                "anchor": 5,
                                "args": [5, 4],
                                "function": "subtotal",
                                "name": "Conservative net",
                            },
                            {
                                "anchor": "fake anchor",
                                "args": ["fake_arg_1", "fake_arg_2"],
                                "function": "fake_fcn_name_not_subtotal",
                                "name": "Fake Name",
                            },
                        ]
                    }
                }
            },
            "type": {
                "categories": [
                    {
                        "numeric_value": 1,
                        "id": 1,
                        "name": "President Obama",
                        "missing": False,
                    },
                    {
                        "numeric_value": 2,
                        "id": 2,
                        "name": "Republicans in Congress",
                        "missing": False,
                    },
                    {"numeric_value": 5, "id": 5, "name": "Not sure", "missing": False},
                    {"numeric_value": 4, "id": 4, "name": "Neither", "missing": False},
                ],
                "class": "categorical",
                "ordinal": False,
            },
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        hs_indices = dimension.hs_indices

        print("hs_indices == %s" % [hs_indices])
        assert hs_indices == ((1, (0, 1)), (2, (2, 3)))

    def test_subtotals(self):
        dimension_dict = {
            "references": {
                "view": {
                    "transform": {
                        "insertions": [
                            {"anchor": 101, "name": "This is respondent ideology"},
                            {
                                "anchor": 2,
                                "args": [1, 2],
                                "function": "subtotal",
                                "name": "Liberal net",
                            },
                            {
                                "anchor": 5,
                                "args": [5, 4],
                                "function": "subtotal",
                                "name": "Conservative net",
                            },
                            {
                                "anchor": "fake anchor",
                                "args": ["fake_arg_1", "fake_arg_2"],
                                "function": "fake_fcn_name_not_subtotal",
                                "name": "Fake Name",
                            },
                        ]
                    }
                }
            },
            "type": {
                "categories": [{"id": 1}, {"id": 5}, {"id": 8}, {"id": 9}, {"id": -1}],
                "class": "categorical",
            },
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        subtotals = dimension._subtotals

        assert len(subtotals) == 2

        subtotal = subtotals[0]
        assert isinstance(subtotal, _Subtotal)
        assert subtotal.anchor == "bottom"
        assert subtotal.addend_ids == (1,)
        assert subtotal.addend_idxs == (0,)
        assert subtotal.label == "Liberal net"

        subtotal = subtotals[1]
        assert isinstance(subtotal, _Subtotal)
        assert subtotal.anchor == 5
        assert subtotal.addend_ids == (5,)
        assert subtotal.addend_idxs == (1,)
        assert subtotal.label == "Conservative net"

    def test_numeric_values(self):
        dimension_dict = {
            "type": {
                "categories": [
                    {"id": 42, "missing": False, "numeric_value": 1},
                    {"id": 43, "missing": False, "numeric_value": 2},
                    {"id": 44, "missing": True, "numeric_value": 3},
                    {"id": 45, "missing": False, "numeric_value": None},
                    {"id": 46, "missing": False},
                ],
                "class": "categorical",
            }
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        numeric_values = dimension.numeric_values

        self.assertEqual(numeric_values, (1, 2, np.nan, np.nan))



# encoding: utf-8
# pylint: disable=too-few-public-methods, invalid-name

"""The CDF of the largest eigenvalue of a Wishart distribution.

Compute the CDF of the Wishart distribution using the algorithm of
Chiani (2014). The Wishart CDF is not defined in scipy.stats.wishart, nor in
tensorflow_probability.distributions.wishart. At the time of writing, the only
implementation of this CDF is the R package rootWishart. This largely
follows that, but using scipy implementations rather than Boost, and omitting
multi-precision support.

For more info about the formulae, and the terse variable naming,
please refer to the paper:

"Distribution of the largest eigenvalue for real Wishart and Gaussian random
matrices and a simple approximation for the Tracy-Widom distribution"

found at: https://arxiv.org/pdf/1209.3394.pdf
"""

from __future__ import division

import numpy as np
from scipy.special import gamma, gammainc
from scipy.linalg import det
from scipy.constants import pi
from cr.cube.util import lazyproperty

try:
    xrange
except NameError:  # pragma: no cover
    xrange = range


class WishartCDF:
    """Implementation of Cumulative Distribution Function (CDF)."""

    def __init__(self, chisq, n_min, n_max):
        self._chisq = chisq
        self.n_min = n_min
        self._n_max = n_max

    @lazyproperty
    def values(self):
        """ndarray of wishart CDF values"""
        return self.K * self.wishart_pfaffian

    @lazyproperty
    def wishart_pfaffian(self):
        """ndarray of wishart pfaffian CDF, before normalization"""
        return np.array(
            [Pfaffian(self, val).value for i, val in np.ndenumerate(self._chisq)]
        ).reshape(self._chisq.shape)

    @lazyproperty
    def alpha(self):
        """int representing common basis of off-diagonal elements of A"""
        return 0.5 * (self._n_max - self.n_min - 1)

    @lazyproperty
    def alpha_ind(self):
        """indices of off-diagonal elements of A"""
        return np.arange(self.n_min)

    @lazyproperty
    def alpha_last(self):
        """elements of row and column to append to A when its order/size is odd"""
        return (self._n_max + self.n_min + 1) / 2

    @lazyproperty
    def alpha_vec(self):
        """elements of row and column to append to A when its order/size is odd"""
        return self.alpha_ind + self.alpha + 1

    @lazyproperty
    def other_ind(self):
        """last row or column of square A"""
        return np.full(self.n_min, self.size - 1, dtype=np.int)

    @staticmethod
    def _mgamma(x, m):
        mgamma = np.float_power(pi, 0.25 * m * (m - 1))
        for i in xrange(m):
            mgamma *= gamma(x - 0.5 * i)
        return mgamma

    @lazyproperty
    def K(self):
        """Normalizing constant for wishart CDF."""
        K1 = np.float_power(pi, 0.5 * self.n_min * self.n_min)
        K1 /= (
            np.float_power(2, 0.5 * self.n_min * self._n_max)
            * self._mgamma(0.5 * self._n_max, self.n_min)
            * self._mgamma(0.5 * self.n_min, self.n_min)
        )

        K2 = np.float_power(
            2, self.alpha * self.size + 0.5 * self.size * (self.size + 1)
        )
        for i in xrange(self.size):
            K2 *= gamma(self.alpha + i + 1)
        return K1 * K2

    @lazyproperty
    def size(self):
        """int representing the size of the wishart matrix.

        If the original size is even, that this is the size that's returned. If the
        size is odd, then the next even number is returned, because of the row/col
        insertion, that happens as the part of the algorithm
        """
        return self.n_min + (self.n_min % 2)


class Pfaffian:
    """Implementation of the Pfaffian value object class."""

    def __init__(self, wishart_cdf, chisq_ordinal):
        self._wishart_cdf = wishart_cdf
        self._chisq_val = chisq_ordinal

    @lazyproperty
    def value(self):
        """return float Cumulative Distribution Function.

        The return value represents a floating point number of the CDF of the
        largest eigenvalue of a Wishart(n, p) evaluated at chisq_val.
        """
        wishart = self._wishart_cdf

        # Prepare variables for integration algorithm
        A = self.A
        p = self._gammainc_a
        g = gamma(wishart.alpha_vec)
        q_ind = np.arange(2 * wishart.n_min - 2)
        q_vec = 2 * wishart.alpha + q_ind + 2
        q = np.float_power(0.5, q_vec) * gamma(q_vec) * gammainc(q_vec, self._chisq_val)

        # Perform integration (i.e. calculate Pfaffian CDF)
        for i in xrange(wishart.n_min):
            # TODO consider index tricks instead of iteration here
            b = 0.5 * p[i] * p[i]
            for j in xrange(i, wishart.n_min - 1):
                b -= q[i + j] / (g[i] * g[j + 1])
                A[j + 1, i] = p[i] * p[j + 1] - 2 * b
                A[i, j + 1] = -A[j + 1, i]

        if np.any(np.isnan(A)):
            return 0
        return np.sqrt(det(A))

    @lazyproperty
    def A(self):
        """ndarray - a skew-symmetric matrix for integrating the target distribution"""
        wishart = self._wishart_cdf

        base = np.zeros([wishart.size, wishart.size])
        if wishart.n_min % 2:
            # If matrix has odd number of elements, we need to append a
            # row and a col, in order for the pfaffian algorithm to work
            base = self._make_size_even(base)
        return base

    @lazyproperty
    def _gammainc_a(self):
        return gammainc(self._wishart_cdf.alpha_vec, 0.5 * self._chisq_val)

    def _make_size_even(self, base):
        wishart = self._wishart_cdf
        alpha_ind = wishart.alpha_ind
        other_ind = wishart.other_ind
        alpha_last = wishart.alpha_last

        base[other_ind, alpha_ind] = (
            np.float_power(2, -alpha_last) * self._gammainc_a / gamma(alpha_last)
        )
        base[alpha_ind, other_ind] = -base[other_ind, alpha_ind]

        return base



import numpy as np

from cr.cube.frozen_cube import FrozenCube

from ..fixtures import CR, SM  # ---mnemonic: SM = 'scale means'---


def test_ca_cat_x_items():
    slice_ = FrozenCube(SM.CA_CAT_X_ITEMS).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.50454821, 3.11233766, 3.35788192, 3.33271833]
    )
    assert slice_.scale_means_column is None
    assert slice_.scale_means_column_margin is None


def test_ca_items_x_cat():
    slice_ = FrozenCube(SM.CA_ITEMS_X_CAT).slices[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.50454821, 3.11233766, 3.35788192, 3.33271833]
    )
    assert slice_.scale_means_row_margin is None


def test_ca_x_mr():
    slice_ = FrozenCube(SM.CA_X_MR).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.29787234, 1.8, 1.48730964, np.nan]
    )
    assert slice_.scale_means_column is None
    assert slice_.scale_means_column_margin is None
    assert slice_.scale_means_row_margin == 1.504548211036992

    slice_ = FrozenCube(SM.CA_X_MR).slices[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [3.31746032, 3.10743802, 3.09976976, np.nan]
    )
    assert slice_.scale_means_column is None

    slice_ = FrozenCube(SM.CA_X_MR).slices[2]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [3.31205674, 3.23913043, 3.37745455, np.nan]
    )
    assert slice_.scale_means_column is None

    slice_ = FrozenCube(SM.CA_X_MR).slices[3]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [3.53676471, 3.34814815, 3.3147877, np.nan]
    )
    assert slice_.scale_means_column is None


def test_cat_x_ca_cat_x_items():
    slice_ = FrozenCube(SM.CAT_X_CA_CAT_X_ITEMS).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.34545455, 2.46938776, 2.7037037, 2.65454545]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(SM.CAT_X_CA_CAT_X_ITEMS).slices[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.41935484, 3.25663717, 3.48, 3.58536585]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(SM.CAT_X_CA_CAT_X_ITEMS).slices[2]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.49429038, 3.44905009, 3.59344262, 3.53630363]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(SM.CAT_X_CA_CAT_X_ITEMS).slices[3]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.43365696, 3.02816901, 3.37987013, 3.32107023]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(SM.CAT_X_CA_CAT_X_ITEMS).slices[4]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.22670025, 2.49473684, 2.79848866, 2.78987342]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(SM.CAT_X_CA_CAT_X_ITEMS).slices[5]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.53061224, 3.68421053, 3.9862069, 4.03472222]
    )
    assert slice_.scale_means_column is None


def test_cat_x_cat():
    slice_ = FrozenCube(SM.CAT_X_CAT).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.6009281, 2.3522267, 2.3197279, 3.3949192]
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [1.43636364, 2.45238095, 2.4730832, 2.68387097, 2.8375, 2.15540541],
    )

    # Test ScaleMeans marginal
    assert slice_.scale_means_column_margin == 2.536319612590799
    assert slice_.scale_means_row_margin == 2.6846246973365617


def test_cat_x_mr():
    slice_ = FrozenCube(SM.CAT_X_MR).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.45070423, 2.54471545, 2.54263006, np.nan]
    )
    assert slice_.scale_means_column is None

    assert slice_.scale_means_column_margin is None
    assert slice_.scale_means_row_margin == 2.5323565323565322


def test_mr_x_cat():
    slice_ = FrozenCube(SM.MR_X_CAT).slices[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [2.45070423, 2.54471545, 2.54263006, np.nan]
    )

    assert slice_.scale_means_column_margin == 2.5323565323565322
    assert slice_.scale_means_row_margin is None


def test_univariate_cat():
    slice_ = FrozenCube(SM.UNIVARIATE_CAT).slices[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [2.6865854])


def test_cat_x_cat_with_hs():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = FrozenCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS, transforms=transforms).slices[
        0
    ]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85],
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [3.87368421, 2.51767677, 3.38429752, 3.66666667, 4.13235294],
    )

    # Test with H&S
    slice_ = FrozenCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).slices[0]
    slice_.scale_means_row
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85],
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [3.87368421, 2.51767677, 3.0851689, 3.38429752, 3.66666667, 4.13235294],
    )


def test_univariate_with_hs():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = FrozenCube(CR.ECON_BLAME_WITH_HS, transforms).slices[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [2.17352056])

    # Test with H&S
    slice_ = FrozenCube(CR.ECON_BLAME_WITH_HS).slices[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [2.17352056])


def test_cat_x_cat_with_hs_on_both_dims():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = FrozenCube(
        CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS, transforms=transforms
    ).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85],
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [3.87368421, 2.51767677, 3.38429752, 3.66666667, 4.13235294],
    )

    # Test with H&S
    slice_ = FrozenCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 2.2423698, 1.88990826, 1.76363636, 3.85],
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [3.87368421, 2.51767677, 3.0851689, 3.38429752, 3.66666667, 4.13235294],
    )


def test_ca_x_mr_with_hs_and_pruning():
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = FrozenCube(CR.CA_X_MR_HS, transforms=transforms).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.50818336, 2.56844883, 2.90251939, np.nan]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(CR.CA_X_MR_HS, transforms=transforms).slices[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.78385708, 2.69292009, 3.11594714, np.nan]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(CR.CA_X_MR_HS, transforms=transforms).slices[2]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [np.nan, np.nan, np.nan, np.nan]
    )
    assert slice_.scale_means_column is None

    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = FrozenCube(CR.CA_X_MR_HS, transforms=transforms).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.50818336, 2.56844883, 2.90251939]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(CR.CA_X_MR_HS, transforms=transforms).slices[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.78385708, 2.69292009, 3.11594714]
    )
    assert slice_.scale_means_column is None
    slice_ = FrozenCube(CR.CA_X_MR_HS, transforms=transforms).slices[2]
    np.testing.assert_almost_equal(slice_.scale_means_row, [])
    assert slice_.scale_means_column is None


def test_cat_x_cat_pruning_and_hs():
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = FrozenCube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.57933884, 2.10618401, 2.30460074, np.nan, 2.34680135]
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [1.74213625, 1.97, 2.45356177, 2.11838791, np.nan, 2.0],
    )

    # Just H&S
    slice_ = FrozenCube(CR.CAT_X_CAT_PRUNING_HS).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [1.57933884, 1.8308135, 2.10618401, 2.30460074, np.nan, 2.34680135],
    ),
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [1.74213625, 2.2364515, 1.97, 2.45356177, 2.11838791, np.nan, 2.0],
    )

    # Just pruning
    transforms = {
        "rows_dimension": {"insertions": {}, "prune": True},
        "columns_dimension": {"insertions": {}, "prune": True},
    }
    slice_ = FrozenCube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.57933884, 2.10618401, 2.30460074, 2.34680135]
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.74213625, 1.97, 2.45356177, 2.11838791, 2.0]
    )

    # Pruning and H&S
    transforms = {
        "rows_dimension": {"insertions": {}, "prune": True},
        "columns_dimension": {"insertions": {}, "prune": True},
    }
    slice_ = FrozenCube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.57933884, 2.106184, 2.3046007, 2.34680135]
    ),
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.74213625, 1.97, 2.45356177, 2.11838791, 2.0]
    )


def test_cat_x_cat_scale_means_margin():
    slice_ = FrozenCube(SM.CAT_X_CAT_SM_MARGIN).slices[0]
    assert slice_.scale_means_row_margin == 2.6846246973365617
    assert slice_.scale_means_column_margin == 2.536319612590799


def test_cat_single_element_x_cat():
    slice_ = FrozenCube(SM.CAT_SINGLE_ELEMENT_X_CAT).slices[0]
    np.testing.assert_equal(slice_.scale_means_row, [np.nan, np.nan, np.nan, np.nan])
    np.testing.assert_equal(slice_.scale_means_column, [np.nan])


def test_means_univariate_cat():
    slice_ = FrozenCube(CR.ECON_BLAME_WITH_HS).slices[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [2.1735205616850553])


def test_means_bivariate_cat():
    slice_ = FrozenCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85],
    )


def test_means_cat_x_mr():
    slice_ = FrozenCube(CR.FRUIT_X_PETS).slices[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.7, 1.6470588, 1.6842105])
    assert slice_.scale_means_column is None


def test_means_mr_x_cat():
    slice_ = FrozenCube(CR.PETS_X_FRUIT).slices[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.7, 1.6470588, 1.6842105]
    )


def test_means_cat_array_cat_dim_first():
    slice_ = FrozenCube(CR.PETS_ARRAY_CAT_FIRST).slices[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.44333002, 1.48049069, 1.57881177]
    )


def test_means_cat_array_subvar_dim_first():
    slice_ = FrozenCube(CR.PETS_ARRAY_SUBVAR_FIRST).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.44333002, 1.48049069, 1.57881177]
    )
    assert slice_.scale_means_column is None


def test_means_cat_x_cat_arr_fruit_first():
    slice_ = FrozenCube(CR.FRUIT_X_PETS_ARRAY).slices[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.48, 1.4285714, 1.5217391]
    )
    slice_ = FrozenCube(CR.FRUIT_X_PETS_ARRAY).slices[1]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.40740741, 1.53846154, 1.55319149]
    )


def test_means_cat_x_cat_arr_subvars_first():
    slice_ = FrozenCube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).slices[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.71111111, 1.6, 1.65625])
    assert slice_.scale_means_column is None

    slice_ = FrozenCube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).slices[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.64705882, 1.7, 1.68421053]
    )
    assert slice_.scale_means_column is None


def test_means_cat_x_cat_arr_pets_first():
    slice_ = FrozenCube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).slices[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.48, 1.40740741])
    np.testing.assert_almost_equal(slice_.scale_means_column, [1.71111111, 1.64705882])

    slice_ = FrozenCube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).slices[1]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.42857143, 1.53846154])
    np.testing.assert_almost_equal(slice_.scale_means_column, [1.6, 1.7])

    slice_ = FrozenCube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).slices[2]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.52173913, 1.55319149])
    np.testing.assert_almost_equal(slice_.scale_means_column, [1.65625, 1.68421053])


def test_means_with_null_values():
    slice_ = FrozenCube(CR.SCALE_WITH_NULL_VALUES).slices[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.2060688, 1.0669344, 1.023199]
    )
    assert slice_.scale_means_column is None



# encoding: utf-8

from __future__ import division

import numpy as np

from cr.cube.dimension import NewDimension
from cr.cube.enum import DIMENSION_TYPE as DT
from cr.cube.frozen_min_base_size_mask import MinBaseSizeMask
from cr.cube.measures.new_pairwise_significance import NewPairwiseSignificance
from cr.cube.matrix import (
    _CatMatrix1D,
    _CatXCatMeansMatrix,
    MatrixFactory,
    MatrixWithHidden,
    MatrixWithInsertions,
    _MeansMatrix0D,
    _MeansMatrix1D,
    _MrMatrix1D,
    _MrWithMeansMatrix1D,
    OrderedMatrix,
    PrunedMatrix,
)
from cr.cube.util import lazyproperty
from cr.cube.vector import _InsertionColumn, _InsertionRow


class FrozenSlice(object):
    """Main point of interaction with the outer world."""

    def __init__(
        self,
        cube,
        slice_idx=0,
        transforms=None,
        population=None,
        ca_as_0th=None,
        mask_size=0,
    ):
        self._cube = cube
        self._slice_idx = slice_idx
        self._transforms_dict = {} if transforms is None else transforms
        self._population = population
        self._ca_as_0th = ca_as_0th
        self._mask_size = mask_size

    # ---interface ---------------------------------------------------

    @lazyproperty
    def _weighted(self):
        return self._cube.is_weighted

    @lazyproperty
    def base_counts(self):
        return np.array([row.base_values for row in self._assembler.rows])

    @lazyproperty
    def column_base(self):
        return np.array([column.base for column in self._assembler.columns]).T

    @lazyproperty
    def column_index(self):
        """ndarray of column index percentages.

        The index values represent the difference of the percentages to the
        corresponding baseline values. The baseline values are the univariate
        percentages of the corresponding variable.
        """
        return np.array([row.column_index for row in self._assembler.rows])

    @lazyproperty
    def column_labels(self):
        """Sequence of str column element names suitable for use as column headings."""
        return tuple(column.label for column in self._assembler.columns)

    @lazyproperty
    def column_margin(self):
        return np.array([column.margin for column in self._assembler.columns]).T

    @lazyproperty
    def column_percentages(self):
        return self.column_proportions * 100

    @lazyproperty
    def column_proportions(self):
        return np.array([col.proportions for col in self._assembler.columns]).T

    @lazyproperty
    def columns_dimension_name(self):
        """str name assigned to columns-dimension.

        The empty string ("") for a 0D or 1D slice (until we get to all slices being
        2D). Reflects the resolved dimension-name transform cascade.
        """
        if len(self.dimensions) < 2:
            return ""
        return self.dimensions[1].name

    @lazyproperty
    def columns_dimension_type(self):
        """Member of `cr.cube.enum.DIMENSION_TYPE` describing columns dimension.

        This value is None for a slice with fewer than two dimensions.
        """
        if len(self.dimensions) < 2:
            return None
        return self.dimensions[1].dimension_type

    @lazyproperty
    def counts(self):
        return np.array([row.values for row in self._assembler.rows])

    @lazyproperty
    def dimension_types(self):
        """Sequence of member of `cr.cube.enum.DIMENSION_TYPE` for each dimension.

        Items appear in rows-dimension, columns-dimension order, although there will be
        fewer than two for a slice with less than two dimensions.
        """
        return tuple(dimension.dimension_type for dimension in self.dimensions)

    @lazyproperty
    def insertion_columns_idxs(self):
        return tuple(
            i for i, column in enumerate(self._assembler.columns) if column.is_insertion
        )

    @lazyproperty
    def insertion_rows_idxs(self):
        return tuple(
            i for i, row in enumerate(self._assembler.rows) if row.is_insertion
        )

    @lazyproperty
    def means(self):
        if type(self._assembler._matrix) is _MeansMatrix0D:
            return self._assembler._matrix.means
        return np.array([row.means for row in self._assembler.rows])

    @lazyproperty
    def min_base_size_mask(self):
        return MinBaseSizeMask(self, self._mask_size)

    @lazyproperty
    def name(self):
        return self.rows_dimension_name

    @lazyproperty
    def names(self):
        return self._matrix.names

    @lazyproperty
    def ndim(self):
        _1D_matrix_types = (
            _MrWithMeansMatrix1D,
            _MeansMatrix1D,
            _CatMatrix1D,
            _MrMatrix1D,
        )
        if isinstance(self._matrix, _CatXCatMeansMatrix):
            return 2
        if isinstance(self._matrix, _1D_matrix_types):
            return 1
        elif isinstance(self._matrix, _MeansMatrix0D):
            return 0
        return 2

    @lazyproperty
    def pairwise_indices(self):
        alpha = self._transforms_dict.get("pairwise_indices", {}).get("alpha", 0.05)
        only_larger = self._transforms_dict.get("pairwise_indices", {}).get(
            "only_larger", True
        )
        return NewPairwiseSignificance(
            self, alpha=alpha, only_larger=only_larger
        ).pairwise_indices

    @lazyproperty
    def pairwise_significance_tests(self):
        """tuple of _ColumnPairwiseSignificance tests.

        Result has as many elements as there are columns in the slice. Each
        significance test contains `p_vals` and `t_stats` (ndarrays that represent
        probability values and statistical scores).
        """
        return tuple(
            NewPairwiseSignificance(self).values[column_idx]
            for column_idx in range(len(self._assembler.columns))
        )

    @lazyproperty
    def population_counts(self):
        return (
            self.table_proportions * self._population * self._cube.population_fraction
        )

    @lazyproperty
    def pvals(self):
        return np.array([row.pvals for row in self._assembler.rows])

    @lazyproperty
    def row_base(self):
        return np.array([row.base for row in self._assembler.rows])

    @lazyproperty
    def row_labels(self):
        return tuple(row.label for row in self._assembler.rows)

    @lazyproperty
    def row_margin(self):
        return np.array([row.margin for row in self._assembler.rows])

    @lazyproperty
    def row_percentages(self):
        return self.row_proportions * 100

    @lazyproperty
    def row_proportions(self):
        return np.array([row.proportions for row in self._assembler.rows])

    @lazyproperty
    def rows_dimension_description(self):
        """str description assigned to rows-dimension.

        The empty string ("") for a 0D slice (until we get to all slices being 2D).
        Reflects the resolved dimension-description transform cascade.
        """
        if len(self.dimensions) == 0:
            return ""
        return self.dimensions[0].description

    @lazyproperty
    def rows_dimension_fills(self):
        """sequence of RGB str like "#def032" fill colors for row elements.

        The values reflect the resolved element-fill transform cascade. The length and
        ordering of the sequence correspond to the rows in the slice, including
        accounting for insertions and hidden rows.
        """
        return tuple(row.fill for row in self._assembler.rows)

    @lazyproperty
    def rows_dimension_name(self):
        """str name assigned to rows-dimension.

        The empty string ("") for a 0D slice (until we get to all slices being 2D).
        Reflects the resolved dimension-name transform cascade.
        """
        if len(self.dimensions) == 0:
            return ""
        return self.dimensions[0].name

    @lazyproperty
    def rows_dimension_type(self):
        """Member of DIMENSION_TYPE enum describing type of rows dimension, or None.

        This value is None for a 0D slice (until we get to all slices being 2D).
        """
        if len(self.dimensions) == 0:
            return None
        return self.dimensions[0].dimension_type

    @lazyproperty
    def scale_means_column(self):
        if np.all(np.isnan(self._columns_dimension_numeric)):
            return None

        inner = np.nansum(self._columns_dimension_numeric * self.counts, axis=1)
        not_a_nan_index = ~np.isnan(self._columns_dimension_numeric)
        denominator = np.sum(self.counts[:, not_a_nan_index], axis=1)
        return inner / denominator

    @lazyproperty
    def scale_means_column_margin(self):
        if np.all(np.isnan(self._columns_dimension_numeric)):
            return None

        column_margin = self.column_margin
        if len(column_margin.shape) > 1:
            # Hack for MR, where column margin is a table. Figure how to
            # fix with subclassing
            column_margin = column_margin[0]

        not_a_nan_index = ~np.isnan(self._columns_dimension_numeric)
        return np.nansum(self._columns_dimension_numeric * column_margin) / np.sum(
            column_margin[not_a_nan_index]
        )

    @lazyproperty
    def scale_means_row(self):
        if np.all(np.isnan(self._rows_dimension_numeric)):
            return None
        inner = np.nansum(self._rows_dimension_numeric[:, None] * self.counts, axis=0)
        not_a_nan_index = ~np.isnan(self._rows_dimension_numeric)
        denominator = np.sum(self.counts[not_a_nan_index, :], axis=0)
        return inner / denominator

    @lazyproperty
    def scale_means_row_margin(self):
        if np.all(np.isnan(self._rows_dimension_numeric)):
            return None

        row_margin = self.row_margin
        if len(row_margin.shape) > 1:
            # Hack for MR, where row margin is a table. Figure how to
            # fix with subclassing
            row_margin = row_margin[:, 0]

        not_a_nan_index = ~np.isnan(self._rows_dimension_numeric)
        return np.nansum(self._rows_dimension_numeric * row_margin) / np.sum(
            row_margin[not_a_nan_index]
        )

    @lazyproperty
    def shape(self):
        return self.counts.shape

    @lazyproperty
    def summary_pairwise_indices(self):
        alpha = self._transforms_dict.get("pairwise_indices", {}).get("alpha", 0.05)
        only_larger = self._transforms_dict.get("pairwise_indices", {}).get(
            "only_larger", True
        )
        return NewPairwiseSignificance(
            self, alpha=alpha, only_larger=only_larger
        ).summary_pairwise_indices

    @lazyproperty
    def table_base(self):
        return self._assembler.table_base

    @lazyproperty
    def table_base_unpruned(self):
        return self._assembler.table_base_unpruned

    @lazyproperty
    def table_margin(self):
        return self._assembler.table_margin

    @lazyproperty
    def table_margin_unpruned(self):
        return self._assembler.table_margin_unpruned

    @lazyproperty
    def table_name(self):
        if self._cube.ndim < 3 and not self._ca_as_0th:
            return None

        title = self._cube.name
        table_name = self._cube.dimensions[0].valid_elements[self._slice_idx].label
        return "%s: %s" % (title, table_name)

    @lazyproperty
    def table_percentages(self):
        return self.table_proportions * 100

    @lazyproperty
    def table_proportions(self):
        return np.array([row.table_proportions for row in self._assembler.rows])

    @lazyproperty
    def zscore(self):
        return np.array([row.zscore for row in self._assembler.rows])

    # ---implementation (helpers)-------------------------------------

    @lazyproperty
    def _assembler(self):
        return _Assembler(self._matrix, self._transforms)

    @lazyproperty
    def _columns_dimension_numeric(self):
        return np.array([column.numeric for column in self._assembler.columns])

    @lazyproperty
    def dimensions(self):
        """tuple of (row,) or (row, col) Dimension objects, depending on 1D or 2D."""
        # TODO: pretty messy while we're shimming in NewDimensions, should clean up
        # pretty naturally after FrozenSlice has its own loader.

        dimensions = self._cube.dimensions[-2:]

        # ---special-case for 0D mean cube---
        if not dimensions:
            return dimensions

        if self._ca_as_0th:
            # Represent CA slice as 1-D rather than 2-D
            dimensions = (dimensions[-1],)

        rows_dimension = NewDimension(
            dimensions[0], self._transforms_dict.get("rows_dimension", {})
        )

        if len(dimensions) == 1:
            return (rows_dimension,)

        columns_dimension = NewDimension(
            dimensions[1], self._transforms_dict.get("columns_dimension", {})
        )

        return (rows_dimension, columns_dimension)

    @lazyproperty
    def _pruning(self):
        """True if any of dimensions has pruning."""
        # TODO: Implement separarte pruning for rows and columns
        return any(dimension.prune for dimension in self.dimensions)

    @lazyproperty
    def _rows_dimension_numeric(self):
        return np.array([row.numeric for row in self._assembler.rows])

    @lazyproperty
    def _matrix(self):
        """The pre-transforms matrix for this slice."""
        cube = self._cube
        base_counts = self._cube.base_counts
        counts_with_missings = self._cube.counts_with_missings
        counts = self._cube.counts
        return MatrixFactory.matrix(
            self.dimensions,
            counts,
            base_counts,
            counts_with_missings,
            cube,
            self._slice_idx,
            self._ca_as_0th,
        )

    @lazyproperty
    def _transforms(self):
        return _Transforms(self._matrix, self.dimensions, self._pruning)


class _Assembler(object):
    """In charge of performing all the transforms sequentially."""

    def __init__(self, matrix, transforms):
        self._matrix = matrix
        self._transforms = transforms

    @lazyproperty
    def columns(self):
        return self.matrix.columns

    @lazyproperty
    def matrix(self):
        """Apply all transforms sequentially."""
        matrix = OrderedMatrix(self._matrix, self._transforms)
        matrix = MatrixWithInsertions(matrix, self._transforms)
        matrix = MatrixWithHidden(matrix, self._transforms)
        matrix = PrunedMatrix(matrix, self._transforms)
        return matrix

    @lazyproperty
    def rows(self):
        return self.matrix.rows

    @lazyproperty
    def table_base(self):
        return self.matrix.table_base

    @lazyproperty
    def table_base_unpruned(self):
        return self.matrix.table_base_unpruned

    @lazyproperty
    def table_margin(self):
        return self.matrix.table_margin

    @lazyproperty
    def table_margin_unpruned(self):
        return self.matrix.table_margin_unpruned


class _Insertions(object):
    """Represents slice's insertions (inserted rows and columns).

    It generates the inserted rows and columns directly from the matrix, based on the
    subtotals.
    """

    def __init__(self, dimensions, matrix):
        self._dimensions = dimensions
        self._matrix = matrix

    @lazyproperty
    def bottom_columns(self):
        return tuple(
            columns for columns in self._inserted_columns if columns.anchor == "bottom"
        )

    @lazyproperty
    def bottom_rows(self):
        return tuple(row for row in self._rows if row.anchor == "bottom")

    @lazyproperty
    def columns(self):
        return tuple(
            column
            for column in self._inserted_columns
            if column.anchor not in ("top", "bottom")
        )

    @lazyproperty
    def rows(self):
        return tuple(row for row in self._rows if row.anchor not in ("top", "bottom"))

    @lazyproperty
    def top_columns(self):
        return tuple(
            columns for columns in self._inserted_columns if columns.anchor == "top"
        )

    @lazyproperty
    def top_rows(self):
        return tuple(row for row in self._rows if row.anchor == "top")

    @lazyproperty
    def _column_dimension(self):
        return self._dimensions[1]

    @lazyproperty
    def _inserted_columns(self):
        """Sequence of _InsertionColumn objects representing subtotal columns."""
        # ---a 1D slice (strand) can have no inserted columns---
        if len(self._dimensions) < 2:
            return ()
        # ---an aggregate columns-dimension is not summable---
        if self._column_dimension.dimension_type in (DT.MR, DT.CA):
            return ()

        return tuple(
            _InsertionColumn(self._matrix, subtotal)
            for subtotal in self._column_dimension.subtotals
        )

    @lazyproperty
    def _row_dimension(self):
        return self._dimensions[0]

    @lazyproperty
    def _rows(self):
        if self._row_dimension.dimension_type in (DT.MR, DT.CA):
            return tuple()

        return tuple(
            _InsertionRow(self._matrix, subtotal)
            for subtotal in self._row_dimension.subtotals
        )


class _OrderTransform(object):
    """Creates ordering indexes for rows and columns based on element ids."""

    def __init__(self, dimensions):
        self._dimensions = dimensions

    @lazyproperty
    def column_order(self):
        """Indexer value identifying columns in order, suitable for slicing an ndarray.

        This value is `slice(None)` when there is no columns dimension. Otherwise it is
        a 1D ndarray of int column indices, suitable for indexing the columns array to
        produce an ordered version.
        """
        # ---if there's no column dimension, there can be no reordering for it---
        if len(self._dimensions) < 2:
            return slice(None)

        # ---Specifying int type prevents failure when there are zero columns. The
        # ---default type for ndarray is float, which is not valid for indexing.
        return np.array(self._columns_dimension.valid_display_order, dtype=int)

    @lazyproperty
    def row_order(self):
        return np.array(self._rows_dimension.valid_display_order, dtype=int)

    @lazyproperty
    def _columns_dimension(self):
        return self._dimensions[1]

    @lazyproperty
    def _rows_dimension(self):
        return self._dimensions[0]


class _Transforms(object):
    """Container for the transforms."""

    def __init__(self, matrix, dimensions, pruning=None):
        self._matrix = matrix
        self._dimensions = dimensions
        self._pruning = pruning

    @lazyproperty
    def insertions(self):
        return _Insertions(self._dimensions, self._matrix)

    @lazyproperty
    def ordering(self):
        return _OrderTransform(self._dimensions)

    @lazyproperty
    def pruning(self):
        return self._pruning



# encoding: utf-8

"""Objects related to "scale-mean" features.

Categorical variables (and perhaps others) allow a *numeric-value* to be
assigned to each category. This could be something like "like=1, dislike=-1,
neutral=0", or possibly "one-child=1, two-children=2, three-or-more=3".

These numeric values impose a quantitative *scale* on an otherwise
non-numeric variable, and thereby allow quantitative operations on that
variable.

Taking the mean value is one of those quantitative operations, and
"scale-mean" refers to calculating the *mean* value of the values on this
numeric *scale*.
"""

from __future__ import division

import numpy as np

from cr.cube.util import lazyproperty
from cr.cube.enum import DIMENSION_TYPE as DT


class ScaleMeans(object):
    """Value object providing "mean-of-scale" values on a CubeSlice object."""

    def __init__(self, slice_):
        self._slice = slice_

    @lazyproperty
    def data(self):
        """list of mean numeric values of categorical responses."""
        means = []
        table = self._slice.as_array()
        products = self._inner_prods(table, self.values)

        for axis, product in enumerate(products):
            if product is None:
                means.append(product)
                continue

            # Calculate means
            valid_indices = self._valid_indices(axis)
            num = np.sum(product[valid_indices], axis)
            den = np.sum(table[valid_indices], axis)
            mean = num / den
            if not isinstance(mean, np.ndarray):
                mean = np.array([mean])
            means.append(mean)
        return means

    def margin(self, axis):
        """Return marginal value of the current slice scaled means.

        This value is the the same what you would get from a single variable
        (constituting a 2D cube/slice), when the "non-missing" filter of the
        opposite variable would be applied. This behavior is consistent with
        what is visible in the front-end client.
        """
        if self._slice.ndim < 2:
            msg = (
                "Scale Means marginal cannot be calculated on 1D cubes, as"
                "the scale means already get reduced to a scalar value."
            )
            raise ValueError(msg)

        dimension_index = 1 - axis
        margin = self._slice.margin(axis=axis)
        if len(margin.shape) > 1:
            index = tuple(
                0 if d.dimension_type == DT.MR else slice(None)
                for d in self._slice.dimensions
            )
            margin = margin[index]
        total = np.sum(margin)
        values = self.values[dimension_index]

        if values is None:
            return None

        return np.sum(values * margin) / total

    def _valid_indices(self, axis):
        # --there is an interaction with CrunchCube._fix_shape() which
        # --essentially eliminates length-1 dimensions not in the first
        # --position. We must mirror that reshaping here. The fact this logic
        # --needs to be duplicated indicates we're missing an abstraction
        # --somewhere, like perhaps CrunchCube and/or
        # --CrunchSlice.reshaped_dimensions.
        reshaped_dimensions = [
            dimension
            for i, dimension in enumerate(self._slice.dimensions)
            if len(dimension.numeric_values) > 1 or i == 0
        ]

        return tuple(
            (
                ~np.isnan(np.array(dim.numeric_values))
                if (
                    dim.numeric_values
                    and any(~np.isnan(dim.numeric_values))
                    and axis == i
                )
                else slice(None)
            )
            for i, dim in enumerate(reshaped_dimensions)
        )

    @lazyproperty
    def values(self):
        """list of ndarray value-ids for each dimension in slice.

        The values for each dimension appear as an ndarray. None appears
        instead of the array for each dimension having only NaN values.
        """
        return [
            (
                np.array(dim.numeric_values)
                if (dim.numeric_values and any(~np.isnan(dim.numeric_values)))
                else None
            )
            for dim in self._slice.dimensions
        ]

    def _inflate(self, dim_ind):
        return self._slice.ndim > 1 and len(self._slice.get_shape()) > 1 and not dim_ind

    def _inner_prods(self, contents, values):
        products = []
        for i, numeric in enumerate(values):
            if numeric is None:
                products.append(numeric)
                continue
            numeric = numeric[:, None] if self._inflate(i) else numeric
            product = contents * numeric
            products.append(product)
        return products



"""This module contains tests for the correct index functionality."""

# pylint: disable=missing-docstring, invalid-name

import numpy as np
import pytest

from cr.cube.frozen_cube import FrozenCube

from ..fixtures import CR


def test_cat_x_cat_slice_column_index():
    slice_ = FrozenCube(CR.CAT_X_CAT).slices[0]
    expected = np.array([[100, 80], [100, 120]])
    np.testing.assert_almost_equal(slice_.column_index, expected)


@pytest.mark.xfail(reason="Row index not implemented in frozen slice", strict=True)
def test_cat_x_cat_slice_row_index():
    slice_ = FrozenCube(CR.CAT_X_CAT).slices[0]
    expected = np.array([[107.142857142857, 85.7142857142857], [93.75, 112.5]])
    np.testing.assert_almost_equal(slice_.row_index, expected)


@pytest.mark.xfail(reason="Row index not implemented in frozen slice", strict=True)
def test_cat_x_cat_slice_row_index_with_baseline():
    slice_ = FrozenCube(CR.CAT_X_CAT).slices[0]
    expected = np.array(
        [[119.047619047619, 71.4285714285714], [104.16666666666667, 93.75]]
    )
    # actual = cube.slices[0].index_table(axis=1, baseline=np.array([0.6, 0.4]))
    np.testing.assert_almost_equal(slice_.row_index, expected)


@pytest.mark.xfail(
    reason="Custom baseline not implemented in frozen slice", strict=True
)
def test_cat_x_cat_slice_column_index_with_baseline():
    slice_ = FrozenCube(CR.CAT_X_CAT).slices[0]
    expected = np.array([[83.3333333333333, 66.6666666666667], [125, 150]])
    # actual = cube.slices[0].index_table(axis=0, baseline=np.array([0.6, 0.4]))
    np.testing.assert_almost_equal(slice_.column_index, expected)


def test_mr_x_cat_slice_column_index():
    slice_ = FrozenCube(CR.SELECTED_CROSSTAB_4).slices[0]
    expected = np.array(
        [
            [95.8651525855387, 103.859044435659],
            [102.305106635277, 97.8432045727022],
            [97.6031146323114, 102.24274029149],
            [98.1029444304978, 101.829068203842],
            [114.466510106092, 86.0656684647625],
            [99.2925720053358, 100.682933745397],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)


@pytest.mark.xfail(reason="Row index not implemented in frozen slice", strict=True)
def test_mr_x_cat_slice_row_index():
    slice_ = FrozenCube(CR.SELECTED_CROSSTAB_4).slices[0]
    expected = np.array(
        [
            [95.3416155822363, 104.394053238208],
            [101.879419344372, 98.2272247381305],
            [97.1985863211465, 102.642452778304],
            [99.2098729346168, 100.745292805163],
            [115.700063998356, 85.1908063256891],
            [100.477252947149, 99.5498278652431],
        ]
    )
    np.testing.assert_almost_equal(slice_.row_index, expected)


def test_mr_x_mr_slice_column_index():
    mr_x_mr = FrozenCube(CR.FULL_CUBE).slices[0]
    mr_alone = FrozenCube(CR.MR_WGTD).slices[0]
    expected = mr_x_mr.column_proportions / mr_alone.table_proportions * 100
    np.testing.assert_almost_equal(mr_x_mr.column_index, expected)


@pytest.mark.xfail(reason="Row index not implemented in frozen slice", strict=True)
def test_mr_x_mr_slice_row_index():
    slice_ = FrozenCube(CR.FULL_CUBE).slices[0]
    expected = [
        96.5399786,
        101.08725891,
        89.22034634,
        89.31908705,
        97.03625304,
        99.96205366,
        79.74421663,
        94.32481007,
        98.95581177,
        101.14422043,
        95.50739913,
        91.82091585,
        92.66856944,
        102.10571127,
        94.88279498,
        84.22596655,
        92.62629722,
        86.77934972,
        99.31115914,
        98.72846269,
        99.54678433,
        94.13302782,
        101.99733805,
        102.24392708,
        97.87112979,
        95.08750269,
        100.61288629,
    ]
    np.testing.assert_almost_equal(slice_.row_index, expected)


def test_cat_x_mr_slice_column_index():
    slice_ = FrozenCube(CR.CAT_X_MR).slices[0]
    expected = np.array(
        [[90.9090909, 106.9518717, 95.6937799], [104.4776119, 96.5759438, 102.1209741]]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)


@pytest.mark.xfail(reason="Row index not implemented in frozen slice", strict=True)
def test_cat_x_mr_slice_row_index():
    slice_ = FrozenCube(CR.CAT_X_MR).slices[0]
    expected = np.array(
        [[85.7142857, 111.5294118, 96.1098398], [107.6923077, 94.6623094, 101.9036954]]
    )
    np.testing.assert_almost_equal(slice_.row_index, expected)


def test_mr_x_mr_index_tables_parity_with_whaam_and_r():
    slice_ = FrozenCube(CR.MR_X_MR_INDEX_TABLE).slices[0]
    # Test column direction
    expected = np.array(
        [
            [192.05298013, 97.23165321, 89.68799602],
            [99.22588537, 239.38592924, 95.99096915],
            [93.52597694, 98.08689727, 182.31556654],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)

    # TODO: Implement in frozen slice
    # # Test row direction
    # expected = np.array(
    #     [
    #         [192.0529801, 99.2258854, 93.5259769],
    #         [97.2316532, 239.3859292, 98.0868973],
    #         [89.687996, 95.9909692, 182.3155665],
    #     ]
    # )
    # actual = cat_x_mr.slices[0].index_table(axis=1)
    # np.testing.assert_almost_equal(actual, expected)


def test_mr_x_3vl_index_tables_parity_with_nssat():
    slice_ = FrozenCube(CR.NSSAT_MR_X_3vl).slices[0]
    # Test column direction
    expected = np.array(
        [
            [179.793686976007, 90.4924459426829],
            [182.343952039497, 88.5838171105893],
            [84.6174957937067, 101.993855386627],
            [np.nan, np.nan],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)

    # TODO: Implement in frozen slice
    # # Test row direction
    # expected = np.array(
    #     [
    #         [179.360970521192, 90.2746542556969],
    #         [184.52243729878, 89.6421386925666],
    #         [84.5355650535557, 101.895100017134],
    #         [np.nan, np.nan],
    #     ]
    # )
    # actual = mr_x_3vl.slices[0].index_table(axis=1)
    # np.testing.assert_almost_equal(actual, expected)


def test_mr_x_mr_index_tables_parity_with_nssat():
    slice_ = FrozenCube(CR.NSSAT_MR_X_MR).slices[0]

    # Test column direction
    expected = np.array(
        [
            [
                114.917891097666,
                94.6007480891202,
                75.7981149285497,
                41.5084915084915,
                64.5687645687646,
                581.118881118881,
                np.nan,
                0,
                np.nan,
            ],
            [
                90.0597657183839,
                95.9426026719446,
                102.497687326549,
                84.1945288753799,
                261.93853427896,
                0,
                np.nan,
                0,
                np.nan,
            ],
            [
                99.4879510762734,
                101.567130443518,
                101.446145177951,
                106.834310398025,
                86.4170866330693,
                59.8272138228942,
                np.nan,
                119.654427645788,
                np.nan,
            ],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)

    # TODO: Implement in frozen slice
    # # Test row direction
    # expected = np.array(
    #     [
    #         [
    #             104.349919743178,
    #             85.9011627906977,
    #             68.8276397515528,
    #             37.6913265306122,
    #             58.6309523809524,
    #             527.678571428571,
    #             np.nan,
    #             0,
    #             np.nan,
    #         ],
    #         [
    #             98.1631656082071,
    #             104.575328614762,
    #             111.7202268431,
    #             91.7701863354037,
    #             285.507246376812,
    #             0,
    #             np.nan,
    #             0,
    #             np.nan,
    #         ],
    #         [
    #             99.6740889304191,
    #             101.757158356526,
    #             101.635946732516,
    #             107.03419298754,
    #             86.57876943881,
    #             59.9391480730223,
    #             np.nan,
    #             119.878296146045,
    #             np.nan,
    #         ],
    #         [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
    #     ]
    # )
    # actual = mr_x_mr.slices[0].index_table(axis=1)
    # np.testing.assert_almost_equal(actual, expected)


def test_mr_single_cat_x_mr():
    # No pruning
    slice_ = FrozenCube(CR.MR_SINGLE_CAT_X_MR).slices[0]
    expected = [[100, 100, np.nan]]
    np.testing.assert_array_equal(slice_.column_index, expected)
    # TODO: Implement row index table
    # np.testing.assert_array_equal(cube_slice.index_table(axis=1), expected)

    # With pruning
    transforms = {"rows_dimension": {"prune": True}}
    slice_ = FrozenCube(CR.MR_SINGLE_CAT_X_MR, transforms=transforms).slices[0]
    expected = [[100, 100]]
    np.testing.assert_array_equal(slice_.column_index, expected)


def test_mr_x_mr_single_cat():
    slice_ = FrozenCube(CR.MR_X_MR_SINGLE_CAT).slices[0]
    expected = [[100], [100], [np.nan]]
    np.testing.assert_array_equal(slice_.column_index, expected)

    # With pruning
    transforms = {"rows_dimension": {"prune": True}}
    slice_ = FrozenCube(CR.MR_X_MR_SINGLE_CAT, transforms=transforms).slices[0]
    expected = [[100], [100]]
    np.testing.assert_array_equal(slice_.column_index, expected)



# encoding: utf-8

"""Vector objects used by frozen-slice.

A vector represents a row or column of data in the overall data matrix. It composes the
element that corresponds to the row or column and so knows the name, element_id, numeric
value, etc. for the row or column.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from scipy.stats import norm

from cr.cube.util import lazyproperty


class BaseVector(object):
    """Base class for all vector objects, although is used directly in some cases."""

    def __init__(self, element, base_counts):
        self._element = element
        self._base_counts = base_counts

    @lazyproperty
    def base(self):
        return np.sum(self._base_counts)

    @lazyproperty
    def fill(self):
        """str RGB color like "#def032" or None when not specified.

        The value reflects the resolved element-fill transform cascade. A value of
        `None` indicates no element-fill transform was specified and the default
        (theme-specified) color should be used for this element.
        """
        return self._element.fill

    @lazyproperty
    def hidden(self):
        return self._element.is_hidden

    @lazyproperty
    def is_insertion(self):
        return False

    @lazyproperty
    def label(self):
        return self._element.label

    @lazyproperty
    def numeric(self):
        return self._element.numeric_value

    @lazyproperty
    def pruned(self):
        return self.base == 0 or np.isnan(self.base)


class _BaseInsertionVector(object):
    """Represents constituent vectors of the `Insertions` class.

    Needs to repro the API of the more basic vectors - because of
    composition (and not inheritance)
    """

    def __init__(self, slice_, subtotal):
        self._slice = slice_
        self._subtotal = subtotal

    @lazyproperty
    def addend_idxs(self):
        return np.array(self._subtotal.addend_idxs)

    @lazyproperty
    def anchor(self):
        return self._subtotal.anchor_idx

    @lazyproperty
    def base(self):
        return np.sum(np.array([vec.base for vec in self._addend_vectors]), axis=0)

    @lazyproperty
    def base_values(self):
        return np.sum(
            np.array([row.base_values for row in self._addend_vectors]), axis=0
        )

    @lazyproperty
    def column_index(self):
        # TODO: Calculate insertion column index for real. Check with Mike
        return np.array([np.nan] * len(self.values))

    @lazyproperty
    def fill(self):
        """Unconditionally `None` for an insertion vector.

        A `fill` value is normally a str RGB value like "#da09fc", specifying the color
        to use for a chart category or series representing this element. The value
        reflects the resolved element-fill transform cascade. Since an insertion cannot
        (currently) have a fill-transform, the default value of `None` (indicating "use
        default color") is unconditionally returned.
        """
        return None

    @lazyproperty
    def hidden(self):
        """Insertion cannot be hidden."""
        return False

    @lazyproperty
    def is_insertion(self):
        return True

    @lazyproperty
    def label(self):
        return self._subtotal.label

    @lazyproperty
    def margin(self):
        return np.sum(np.array([vec.margin for vec in self._addend_vectors]), axis=0)

    @lazyproperty
    def means(self):
        return np.array([np.nan])

    @lazyproperty
    def numeric(self):
        return np.nan

    @lazyproperty
    def table_margin(self):
        return self._slice.table_margin

    @lazyproperty
    def values(self):
        return np.sum(np.array([row.values for row in self._addend_vectors]), axis=0)


class CategoricalVector(BaseVector):
    """Main staple of all measures.

    Some of the measures it can calculate by itself, others it needs to receive at
    construction time (like table margin and zscores).
    """

    def __init__(
        self, counts, base_counts, element, table_margin, zscore=None, column_index=None
    ):
        super(CategoricalVector, self).__init__(element, base_counts)
        self._counts = counts
        self._table_margin = table_margin
        self._zscore = zscore
        self._column_index = column_index

    @lazyproperty
    def base_values(self):
        if not isinstance(self._base_counts, np.ndarray):
            return np.array([self._base_counts])
        return self._base_counts

    @lazyproperty
    def column_index(self):
        return self._column_index

    @lazyproperty
    def margin(self):
        return np.sum(self._counts)

    @lazyproperty
    def proportions(self):
        return self.values / self.margin
        # return self.values / self.base

    @lazyproperty
    def pvals(self):
        return 2 * (1 - norm.cdf(np.abs(self._zscore)))

    @lazyproperty
    def table_margin(self):
        return self._table_margin

    @lazyproperty
    def values(self):
        if not isinstance(self._counts, np.ndarray):
            return np.array([self._counts])
        return self._counts

    @lazyproperty
    def zscore(self):
        return self._zscore


class CatXMrVector(CategoricalVector):
    def __init__(
        self, counts, base_counts, label, table_margin, zscore=None, column_index=None
    ):
        super(CatXMrVector, self).__init__(
            counts[0], base_counts[0], label, table_margin, zscore, column_index
        )
        self._pruning_bases = base_counts

    @lazyproperty
    def pruned(self):
        return np.sum(self._pruning_bases) == 0


class _InsertionColumn(_BaseInsertionVector):
    """Represents an inserted (subtotal) column."""

    @lazyproperty
    def pruned(self):
        return not np.any(np.array([row.base for row in self._slice.rows]))

    @lazyproperty
    def _addend_vectors(self):
        return tuple(
            column
            for i, column in enumerate(self._slice.columns)
            if i in self._subtotal.addend_idxs
        )


class _InsertionRow(_BaseInsertionVector):
    """Represents an inserted (subtotal) row."""

    @lazyproperty
    def pruned(self):
        return not np.any(np.array([column.base for column in self._slice.columns]))

    @lazyproperty
    def pvals(self):
        return np.array([np.nan] * len(self._slice.columns))

    @lazyproperty
    def zscore(self):
        return np.array([np.nan] * len(self._slice.columns))

    @lazyproperty
    def _addend_vectors(self):
        return tuple(
            row
            for i, row in enumerate(self._slice.rows)
            if i in self._subtotal.addend_idxs
        )


class MeansVector(BaseVector):
    def __init__(self, element, base_counts, means):
        super(MeansVector, self).__init__(element, base_counts)
        self._means = means

    @lazyproperty
    def means(self):
        return self._means

    @lazyproperty
    def values(self):
        return self._means


class MeansWithMrVector(MeansVector):
    """This is a row of a 1-D MR with Means.

    This vector is special in the sense that it doesn't provide us with the normal base
    (which is selected + not-selected for a 1-D MR *without* means). Instead, it
    calculates the base as *just* the selected, which is the correct base for the 1-D MR
    with means.
    """

    @lazyproperty
    def base(self):
        return np.sum(self._base_counts[0])


class MultipleResponseVector(CategoricalVector):
    """Handles MR vectors (either rows or columns)

    Needs to handle selected and not-selected properly. Consequently, it calculates the
    right margin (for itself), but receives table margin on construction time (from the
    slice).
    """

    @lazyproperty
    def base(self):
        counts = zip(self._selected_unweighted, self._not_selected_unweighted)
        return np.array(
            [selected + not_selected for (selected, not_selected) in counts]
        )

    @lazyproperty
    def base_values(self):
        return self._base_counts[0, :]

    @lazyproperty
    def margin(self):
        counts = zip(self._selected, self._not_selected)
        return np.array(
            [selected + not_selected for (selected, not_selected) in counts]
        )

    @lazyproperty
    def pruned(self):
        return np.all(self.base == 0) or np.all(np.isnan(self.base))

    @lazyproperty
    def values(self):
        return self._selected

    @lazyproperty
    def _not_selected(self):
        return self._counts[1, :]

    @lazyproperty
    def _not_selected_unweighted(self):
        return self._base_counts[1, :]

    @lazyproperty
    def _selected(self):
        return self._counts[0, :]

    @lazyproperty
    def _selected_unweighted(self):
        return self._base_counts[0, :]


class _TransformedVector(object):
    # @lazyproperty
    # def counts(self):
    #     return self._base_vector.counts

    @lazyproperty
    def fill(self):
        """str RGB color like "#def032" or None when not specified.

        The value reflects the resolved element-fill transform cascade. A value of
        `None` indicates no element-fill transform was specified and the default
        (theme-specified) color should be used for this element.
        """
        return self._base_vector.fill

    @lazyproperty
    def hidden(self):
        return self._base_vector.hidden

    @lazyproperty
    def is_insertion(self):
        return self._base_vector.is_insertion

    @lazyproperty
    def label(self):
        return self._base_vector.label

    @lazyproperty
    def pruned(self):
        return self._base_vector.pruned

    @lazyproperty
    def margin(self):
        return self._base_vector.margin

    @lazyproperty
    def means(self):
        return self._base_vector.means

    @lazyproperty
    def numeric(self):
        return self._base_vector.numeric

    # @lazyproperty
    # def table_base(self):
    #     return self._base_vector.table_base

    @lazyproperty
    def table_margin(self):
        return self._base_vector.table_margin

    # @lazyproperty
    # def values(self):
    #     return self.counts


class AssembledVector(_TransformedVector):
    """Vector with base, as well as inserted, elements (of the opposite dimension)."""

    def __init__(self, base_vector, opposite_inserted_vectors):
        self._base_vector = base_vector
        self._opposite_inserted_vectors = opposite_inserted_vectors

    @lazyproperty
    def base(self):
        return self._base_vector.base

    @lazyproperty
    def base_values(self):
        # TODO: Do for real
        return np.array(
            self._top_base_values
            + self._interleaved_base_values
            + self._bottom_base_values
        )

    @lazyproperty
    def column_index(self):
        # return self._base_vector.column_index
        return np.array(
            tuple([np.nan] * len(self._top_values))
            + self._interleaved_column_index
            + tuple([np.nan] * len(self._bottom_values))
        )

    @lazyproperty
    def label(self):
        return self._base_vector.label

    @lazyproperty
    def margin(self):
        return self._base_vector.margin

    @lazyproperty
    def proportions(self):
        # return self.values / self.base
        return self.values / self.margin

    @lazyproperty
    def pruned(self):
        return self._base_vector.pruned

    @lazyproperty
    def pvals(self):
        return np.array(
            tuple([np.nan] * len(self._top_values))
            + self._interleaved_pvals
            + tuple([np.nan] * len(self._bottom_values))
        )

    @lazyproperty
    def table_proportions(self):
        return self.values / self._base_vector.table_margin

    @lazyproperty
    def values(self):
        return np.array(
            self._top_values + self._interleaved_values + self._bottom_values
        )

    @lazyproperty
    def zscore(self):
        return np.array(
            tuple([np.nan] * len(self._top_values))
            + self._interleaved_zscore
            + tuple([np.nan] * len(self._bottom_values))
        )

    @lazyproperty
    def _bottom_base_values(self):
        return tuple(
            np.sum(self._base_vector.base_values[col.addend_idxs])
            for col in self._opposite_inserted_vectors
            if col.anchor == "bottom"
        )

    @lazyproperty
    def _bottom_values(self):
        return tuple(
            np.sum(self._base_vector.values[col.addend_idxs])
            for col in self._opposite_inserted_vectors
            if col.anchor == "bottom"
        )

    @lazyproperty
    def _interleaved_base_values(self):
        base_values = []
        for i in range(len(self._base_vector.base_values)):
            base_values.append(self._base_vector.base_values[i])
            for inserted_vector in self._opposite_inserted_vectors:
                if i == inserted_vector.anchor:
                    insertion_value = np.sum(
                        self._base_vector.base_values[inserted_vector.addend_idxs]
                    )
                    base_values.append(insertion_value)
        return tuple(base_values)

    @lazyproperty
    def _interleaved_column_index(self):
        # TODO: Replace with real column index values from insertions vectors. This
        # should be something like:
        #   col_ind = (ins1.prop + ins2.prop) / (ins1.baseline + ins2.baseline)
        # ask @mike to confirm
        column_index = []
        for i, value in enumerate(self._base_vector.column_index):
            column_index.append(value)
            for inserted_vector in self._opposite_inserted_vectors:
                if i == inserted_vector.anchor:
                    column_index.append(np.nan)
        return tuple(column_index)

    @lazyproperty
    def _interleaved_pvals(self):
        pvals = []
        for i, value in enumerate(self._base_vector.pvals):
            pvals.append(value)
            for inserted_vector in self._opposite_inserted_vectors:
                if i == inserted_vector.anchor:
                    pvals.append(np.nan)
        return tuple(pvals)

    @lazyproperty
    def _interleaved_values(self):
        values = []
        for i in range(len(self._base_vector.values)):
            values.append(self._base_vector.values[i])
            for inserted_vector in self._opposite_inserted_vectors:
                if i == inserted_vector.anchor:
                    insertion_value = np.sum(
                        self._base_vector.values[inserted_vector.addend_idxs]
                    )
                    values.append(insertion_value)
        return tuple(values)

    @lazyproperty
    def _interleaved_zscore(self):
        zscore = []
        for i, value in enumerate(self._base_vector.zscore):
            zscore.append(value)
            for inserted_vector in self._opposite_inserted_vectors:
                if i == inserted_vector.anchor:
                    zscore.append(np.nan)
        return tuple(zscore)

    @lazyproperty
    def _top_base_values(self):
        return tuple(
            np.sum(self._base_vector.base_values[col.addend_idxs])
            for col in self._opposite_inserted_vectors
            if col.anchor == "top"
        )

    @lazyproperty
    def _top_values(self):
        return tuple(
            np.sum(self._base_vector.values[col.addend_idxs])
            for col in self._opposite_inserted_vectors
            if col.anchor == "top"
        )


class AssembledInsertionVector(AssembledVector):
    """Inserted row or col, but with elements from opposite dimension insertions.

    Needs to be subclassed from _AssembledVector, because it needs to provide the
    anchor, in order to know where it (itself) gets inserted.
    """

    @lazyproperty
    def anchor(self):
        return self._base_vector.anchor


class _BasePrunedOrHiddenVector(_TransformedVector):
    def __init__(self, base_vector, opposite_vectors):
        self._base_vector = base_vector
        self._opposite_vectors = opposite_vectors

    @lazyproperty
    def base(self):
        if not isinstance(self._base_vector.base, np.ndarray):
            return self._base_vector.base
        return self._base_vector.base[self._valid_elements_idxs]

    @lazyproperty
    def base_values(self):
        return self._base_vector.base_values[self._valid_elements_idxs]

    @lazyproperty
    def column_index(self):
        return self._base_vector.column_index[self._valid_elements_idxs]

    @lazyproperty
    def margin(self):
        if not isinstance(self._base_vector.margin, np.ndarray):
            return self._base_vector.margin
        return self._base_vector.margin[self._valid_elements_idxs]

    @lazyproperty
    def means(self):
        return self._base_vector.means[self._valid_elements_idxs]

    @lazyproperty
    def proportions(self):
        return self._base_vector.proportions[self._valid_elements_idxs]

    @lazyproperty
    def pvals(self):
        return self._base_vector.pvals[self._valid_elements_idxs]

    @lazyproperty
    def table_proportions(self):
        return self._base_vector.table_proportions[self._valid_elements_idxs]

    @lazyproperty
    def values(self):
        return self._base_vector.values[self._valid_elements_idxs]

    @lazyproperty
    def zscore(self):
        return self._base_vector.zscore[self._valid_elements_idxs]

    @lazyproperty
    def _valid_elements_idxs(self):
        raise NotImplementedError("must be implemented by each subclass")


class HiddenVector(_BasePrunedOrHiddenVector):
    """Vector with elements from the opposide dimensions hidden."""

    @lazyproperty
    def _valid_elements_idxs(self):
        """An 1D ndarray of int idxs of visible values, suitable for array indexing."""
        return np.array(
            [
                index
                for index, opposite_vector in enumerate(self._opposite_vectors)
                if not opposite_vector.hidden
            ],
            dtype=int,
        )


class OrderedVector(_TransformedVector):
    """In charge of indexing elements properly, after ordering transform."""

    def __init__(self, vector, order):
        self._base_vector = vector
        self._order = order

    @lazyproperty
    def base(self):
        return self._base_vector.base

    @lazyproperty
    def base_values(self):
        return self._base_vector.base_values[self.order]

    @lazyproperty
    def column_index(self):
        return self._base_vector.column_index

    @lazyproperty
    def label(self):
        return self._base_vector.label

    @lazyproperty
    def order(self):
        return self._order if self._order is not None else slice(None)

    @lazyproperty
    def pvals(self):
        return self._base_vector.pvals

    @lazyproperty
    def values(self):
        return self._base_vector.values[self.order]

    @lazyproperty
    def zscore(self):
        return self._base_vector.zscore


class PrunedVector(_BasePrunedOrHiddenVector):
    """Vector with elements from the opposide dimensions pruned."""

    @lazyproperty
    def _valid_elements_idxs(self):
        """An 1D ndarray of int idxs of unpruned values, suitable for array indexing."""
        return np.array(
            [
                index
                for index, opposite_vector in enumerate(self._opposite_vectors)
                if not opposite_vector.pruned
            ],
            dtype=int,
        )



# encoding: utf-8

"""Provides the FrozenCube class.

FrozenCube is the main API class for manipulating Crunch.io JSON cube
responses.
"""

from __future__ import division

import json

import numpy as np

from cr.cube.dimension import AllDimensions
from cr.cube.slices import FrozenSlice
from cr.cube.util import lazyproperty
from cr.cube.enum import DIMENSION_TYPE as DT

np.seterr(divide="ignore", invalid="ignore")


class FrozenCube(object):
    def __init__(
        self,
        response,
        transforms=None,
        first_cube_of_tab=False,
        population=None,
        mask_size=0,
    ):
        self._cube_response_arg = response
        self._transforms_dict = {} if transforms is None else transforms
        self._first_cube_of_tab = first_cube_of_tab
        self._population = 0 if population is None else population
        self._mask_size = mask_size

    def __repr__(self):
        """Provide text representation suitable for working at console.

        Falls back to a default repr on exception, such as might occur in
        unit tests where object need not otherwise be provided with all
        instance variable values.
        """
        try:
            dimensionality = " x ".join(dt.name for dt in self.dimension_types)
            return "%s(name='%s', dimension_types='%s')" % (
                type(self).__name__,
                self.name,
                dimensionality,
            )
        except Exception:
            return super(FrozenCube, self).__repr__()

    @lazyproperty
    def slices(self):
        return tuple(
            FrozenSlice(
                self,
                slice_idx=slice_idx,
                transforms=self._transforms_dict,
                population=self._population,
                ca_as_0th=self._ca_as_0th,
                mask_size=self._mask_size,
            )
            for slice_idx in self._slice_idxs
        )

    @lazyproperty
    def _ca_as_0th(self):
        return self._first_cube_of_tab and self.dimension_types[0] == DT.CA

    @lazyproperty
    def _slice_idxs(self):
        if self.ndim < 3 and not self._ca_as_0th:
            return (0,)
        return range(len(self.dimensions[0].valid_elements))

    @lazyproperty
    def base_counts(self):
        return self._measures.unweighted_counts.raw_cube_array[self._valid_idxs]

    @lazyproperty
    def counts_with_missings(self):
        return self._measure(self.is_weighted).raw_cube_array

    @lazyproperty
    def counts(self):
        return self.counts_with_missings[self._valid_idxs]

    @lazyproperty
    def _valid_idxs(self):
        return np.ix_(
            *tuple(d.valid_elements.element_idxs for d in self._all_dimensions)
        )

    @lazyproperty
    def description(self):
        """Return the description of the cube."""
        if not self.dimensions:
            return None
        return self.dimensions[0].description

    @lazyproperty
    def dimension_types(self):
        """Tuple of DIMENSION_TYPE member for each dimension of cube."""
        return tuple(d.dimension_type for d in self.dimensions)

    @lazyproperty
    def dimensions(self):
        """_ApparentDimension object providing access to visible dimensions.

        A cube involving a multiple-response (MR) variable has two dimensions
        for that variable (subvariables and categories dimensions), but is
        "collapsed" into a single effective dimension for cube-user purposes
        (its categories dimension is supressed). This collection will contain
        a single dimension for each MR variable and therefore may have fewer
        dimensions than appear in the cube response.
        """
        return self._all_dimensions.apparent_dimensions

    @lazyproperty
    def has_means(self):
        """True if cube includes a means measure."""
        return self._measures.means is not None

    @lazyproperty
    def is_weighted(self):
        """True if cube response contains weighted data."""
        return self._measures.is_weighted

    @lazyproperty
    def missing(self):
        """Get missing count of a cube."""
        return self._measures.missing_count

    @lazyproperty
    def name(self):
        """Return the name of the cube.

        If the cube has 2 diensions, return the name of the second one. In case
        of a different number of dimensions, default to returning the name of
        the last one. In case of no dimensions, return the empty string.
        """
        if not self.dimensions:
            return None
        return self.dimensions[0].name

    @lazyproperty
    def ndim(self):
        """int count of dimensions for this cube."""
        return len(self.dimensions)

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """
        return self._measures.population_fraction

    @lazyproperty
    def _all_dimensions(self):
        """The AllDimensions object for this cube.

        The AllDimensions object provides access to all the dimensions
        appearing in the cube response, not only apparent dimensions (those
        that appear to a user). It also provides access to
        an _ApparentDimensions object which contains only those user-apparent
        dimensions (basically the categories dimension of each MR
        dimension-pair is suppressed).
        """
        return AllDimensions(dimension_dicts=self._cube_dict["result"]["dimensions"])

    @lazyproperty
    def _cube_dict(self):
        """dict containing raw cube response, parsed from JSON payload."""
        try:
            cube_response = self._cube_response_arg
            # ---parse JSON to a dict when constructed with JSON---
            cube_dict = (
                cube_response
                if isinstance(cube_response, dict)
                else json.loads(cube_response)
            )
            # ---cube is 'value' item in a shoji response---
            return cube_dict.get("value", cube_dict)
        except TypeError:
            raise TypeError(
                "Unsupported type <%s> provided. Cube response must be JSON "
                "(str) or dict." % type(self._cube_response_arg).__name__
            )

    def _measure(self, weighted):
        """_BaseMeasure subclass representing primary measure for this cube.

        If the cube response includes a means measure, the return value is
        means. Otherwise it is counts, with the choice between weighted or
        unweighted determined by *weighted*.

        Note that weighted counts are provided on an "as-available" basis.
        When *weighted* is True and the cube response is not weighted,
        unweighted counts are returned.
        """
        return (
            self._measures.means
            if self._measures.means is not None
            else self._measures.weighted_counts
            if weighted
            else self._measures.unweighted_counts
        )

    @lazyproperty
    def _measures(self):
        """_Measures object for this cube.

        Provides access to unweighted counts, and weighted counts and/or means
        when available.
        """
        return _Measures(self._cube_dict, self._all_dimensions)


class _Measures(object):
    """Provides access to measures contained in cube response."""

    def __init__(self, cube_dict, all_dimensions):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions

    @lazyproperty
    def is_weighted(self):
        """True if weights have been applied to the measure(s) for this cube.

        Unweighted counts are available for all cubes. Weighting applies to
        any other measures provided by the cube.
        """
        cube_dict = self._cube_dict
        if cube_dict.get("query", {}).get("weight") is not None:
            return True
        if cube_dict.get("weight_var") is not None:
            return True
        if cube_dict.get("weight_url") is not None:
            return True
        unweighted_counts = cube_dict["result"]["counts"]
        count_data = cube_dict["result"]["measures"].get("count", {}).get("data")
        if unweighted_counts != count_data:
            return True
        return False

    @lazyproperty
    def means(self):
        """_MeanMeasure object providing access to means values.

        None when the cube response does not contain a mean measure.
        """
        mean_measure_dict = (
            self._cube_dict.get("result", {}).get("measures", {}).get("mean")
        )
        if mean_measure_dict is None:
            return None
        return _MeanMeasure(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows in cube response."""
        if self.means:
            return self.means.missing_count
        return self._cube_dict["result"].get("missing", 0)

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """
        numerator = self._cube_dict["result"].get("filtered", {}).get("weighted_n")
        denominator = self._cube_dict["result"].get("unfiltered", {}).get("weighted_n")
        try:
            return numerator / denominator
        except ZeroDivisionError:
            return np.nan
        except Exception:
            return 1.0

    @lazyproperty
    def unweighted_counts(self):
        """_UnweightedCountMeasure object for this cube.

        This object provides access to unweighted counts for this cube,
        whether or not the cube contains weighted counts.
        """
        return _UnweightedCountMeasure(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def weighted_counts(self):
        """_WeightedCountMeasure object for this cube.

        This object provides access to weighted counts for this cube, if
        available. If the cube response is not weighted, the
        _UnweightedCountMeasure object for this cube is returned.
        """
        return _WeightedCountMeasure(self._cube_dict, self._all_dimensions)


class _BaseMeasure(object):
    """Base class for measure objects."""

    def __init__(self, cube_dict, all_dimensions):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions

    @lazyproperty
    def raw_cube_array(self):
        """Return read-only ndarray of measure values from cube-response.

        The shape of the ndarray mirrors the shape of the (raw) cube
        response. Specifically, it includes values for missing elements, any
        MR_CAT dimensions, and any prunable rows and columns.
        """
        array = np.array(self._flat_values).reshape(self._all_dimensions.shape)
        # ---must be read-only to avoid hard-to-find bugs---
        array.flags.writeable = False
        return array

    @lazyproperty
    def _flat_values(self):
        """Return tuple of mean values as found in cube response.

        This property must be implemented by each subclass.
        """
        raise NotImplementedError("must be implemented by each subclass")


class _MeanMeasure(_BaseMeasure):
    """Statistical mean values from a cube-response."""

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows reflected in response."""
        return self._cube_dict["result"]["measures"]["mean"].get("n_missing", 0)

    @lazyproperty
    def _flat_values(self):
        """Return tuple of mean values as found in cube response.

        Mean data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        return tuple(
            np.nan if type(x) is dict else x
            for x in self._cube_dict["result"]["measures"]["mean"]["data"]
        )


class _UnweightedCountMeasure(_BaseMeasure):
    """Unweighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """tuple of int counts before weighting."""
        return tuple(self._cube_dict["result"]["counts"])


class _WeightedCountMeasure(_BaseMeasure):
    """Weighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """tuple of numeric counts after weighting."""
        return tuple(self._cube_dict["result"]["measures"]["count"]["data"])



# encoding: utf-8

"""P-values of pairwise comparison or columns of a contingency table."""

from __future__ import division

import numpy as np

from cr.cube.enum import DIMENSION_TYPE as DT
from cr.cube.distributions.wishart import WishartCDF
from cr.cube.util import lazyproperty

try:
    xrange
except NameError:  # pragma: no cover
    # pylint: disable=invalid-name
    xrange = range


# pylint: disable=too-few-public-methods
class WishartPairwiseSignificance:
    """Value object providing matrix of pairwise-comparison P-values"""

    def __init__(self, slice_, axis=0, weighted=True, alpha=0.05):
        self._slice = slice_
        self._axis = axis
        self._weighted = weighted
        self._alpha = alpha

    @classmethod
    def pvals(cls, slice_, axis=0, weighted=True):
        """Wishart CDF values for slice columns as square ndarray.

        Wishart CDF (Cumulative Distribution Function) is calculated to determine
        statistical significance of slice columns, in relation to all other columns.
        These values represent the answer to the question "How much is a particular
        column different from each other column in the slice".
        """
        return cls._factory(slice_, axis, weighted).pvals

    @classmethod
    def pairwise_indices(cls, slice_, axis=0, weighted=True):
        return cls._factory(slice_, axis, weighted).pairwise_indices

    def _chi_squared(self, proportions, margin, observed):
        """return ndarray of chi-squared measures for proportions' columns.

        *proportions* (ndarray): The basis of chi-squared calcualations
        *margin* (ndarray): Column margin for proportions (See `def _margin`)
        *observed* (ndarray): Row margin proportions (See `def _observed`)
        """
        n = self._element_count
        chi_squared = np.zeros([n, n])
        for i in xrange(1, n):
            for j in xrange(0, n - 1):
                denominator = 1 / margin[i] + 1 / margin[j]
                chi_squared[i, j] = chi_squared[j, i] = (
                    np.sum(np.square(proportions[:, i] - proportions[:, j]) / observed)
                    / denominator
                )
        return chi_squared

    @staticmethod
    def _pairwise_sig_indices(significance_by_col):
        """Return tuple like (1, 3), (6,), or () identifying pairwise sig cols.

        The *significance_by_col* parameter is a 1D boolean array with one value for
        each column in this crosstab, a True value signifying pairwise significance.
        """
        return tuple(
            i
            for i, is_pairwise_significant in enumerate(significance_by_col)
            if is_pairwise_significant
        )

    def _pvals_from_chi_squared(self, pairwise_chisq):
        """return statistical significance for props' columns.

        *pairwise_chisq* (ndarray) Matrix of chi-squared values (bases for Wishart CDF)
        """
        return self._intersperse_insertion_rows_and_columns(
            1.0 - WishartCDF(pairwise_chisq, self._n_min, self._n_max).values
        )

    @staticmethod
    def _factory(slice_, axis, weighted):
        """return subclass for PairwiseSignificance, based on slice dimension types."""
        if slice_.dim_types[0] == DT.MR_SUBVAR:
            return _MrXCatPairwiseSignificance(slice_, axis, weighted)
        return _CatXCatPairwiseSignificance(slice_, axis, weighted)

    @lazyproperty
    def _insertion_indices(self):
        """Return H&S indices of the pairwise comparison dimension."""
        return self._slice.inserted_hs_indices()[1 - self._axis]

    def _intersperse_insertion_rows_and_columns(self, pairwise_pvals):
        """Return pvals matrix with inserted NaN rows and columns, as numpy.ndarray.

        Each insertion (a header or a subtotal) creates an offset in the calculated
        pvals. These need to be taken into account when converting each pval to a
        corresponding column letter. For this reason, we need to insert an all-NaN
        row and a column at the right indices. These are the inserted indices of each
        insertion, along respective dimensions.
        """
        for i in self._insertion_indices:
            pairwise_pvals = np.insert(pairwise_pvals, i, np.nan, axis=0)
            pairwise_pvals = np.insert(pairwise_pvals, i, np.nan, axis=1)
        return pairwise_pvals

    @lazyproperty
    def _margin(self):
        """Margin for the axis as numpy.ndarray."""
        return self._slice.margin(axis=self._axis)

    @lazyproperty
    def _element_count(self):
        """Number of elements of the dimension opposite to axis, as int."""
        return self._slice.get_shape()[1 - self._axis]

    @lazyproperty
    def _n_max(self):
        """Size (zero based) of the bigger of the two slice's dimension, as int."""
        return max(self._slice.get_shape()) - 1

    @lazyproperty
    def _n_min(self):
        """Size (zero based) of the smaller of the two slice's dimension, as int."""
        return min(self._slice.get_shape()) - 1

    @lazyproperty
    def _opposite_axis_margin(self):
        """ndarray representing margin along the axis opposite of self._axis

        In the process of calculating p-values for the column significance testing we
        need both the margin along the primary axis and the percentage margin along
        the opposite axis.
        """
        off_axis = 1 - self._axis
        return self._slice.margin(axis=off_axis, include_mr_cat=self._include_mr_cat)

    @lazyproperty
    def _proportions(self):
        """ndarray representing slice proportions along correct axis."""
        return self._slice.proportions(
            axis=self._axis, include_mr_cat=self._include_mr_cat
        )


class _CatXCatPairwiseSignificance(WishartPairwiseSignificance):
    """Pairwise significance for CAT x CAT type slices."""

    _include_mr_cat = False

    @lazyproperty
    def pairwise_indices(self):
        return np.array(
            [
                self._pairwise_sig_indices(pvals_row < self._alpha)
                for pvals_row in self.pvals
            ]
        )

    @lazyproperty
    def pvals(self):
        """Square matrix of pairwise Chi-square along axis, as numpy.ndarray."""
        return self._pvals_from_chi_squared(self._pairwise_chisq)

    @lazyproperty
    def _observed(self):
        """Observed marginal proportions, as float."""
        total = self._slice.margin()
        return self._opposite_axis_margin / total

    @lazyproperty
    def _pairwise_chisq(self):
        """Pairwise comparisons (Chi-Square) along axis, as numpy.ndarray.

        Returns a square, symmetric matrix of test statistics for the null
        hypothesis that each vector along *axis* is equal to each other.
        """
        return self._chi_squared(self._proportions, self._margin, self._observed)


class _MrXCatPairwiseSignificance(WishartPairwiseSignificance):
    """Pairwise significance for MR x CAT type slices."""

    _include_mr_cat = True

    @lazyproperty
    def pairwise_indices(self):
        return np.array(
            [
                [
                    self._pairwise_sig_indices(pvals_row < self._alpha)
                    for pvals_row in pvals_matrix
                ]
                for pvals_matrix in self.pvals
            ]
        )

    @lazyproperty
    def pvals(self):
        """Square matrix of pairwise Chi-square along axis, as numpy.ndarray."""
        return [
            self._pvals_from_chi_squared(mr_subvar_chisq)
            for mr_subvar_chisq in self._pairwise_chisq
        ]

    @lazyproperty
    def _pairwise_chisq(self):
        """Pairwise comparisons (Chi-Square) along axis, as numpy.ndarray.

        Returns a list of square and symmetric matrices of test statistics for the null
        hypothesis that each vector along *axis* is equal to each other.
        """
        return [
            self._chi_squared(
                mr_subvar_proportions,
                self._margin[idx],
                self._opposite_axis_margin[idx]
                / np.sum(self._opposite_axis_margin[idx]),
            )
            for (idx, mr_subvar_proportions) in enumerate(self._proportions)
        ]
